from __future__ import annotations

from dataclasses import asdict
from functools import partial
from pathlib import Path
from typing import Iterable, Optional, Union, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp
import optax
import orbax.checkpoint as ocp
from tqdm import tqdm

import lacss.metrics

from lacss.losses import (
    aux_size_loss,
    collaborator_border_loss,
    collaborator_segm_loss,
    lpn_loss,
    segmentation_loss,
)

from ..modules import Lacss, UNet
from ..typing import Array, Optimizer
from xtrain import Trainer, JIT
from ..typing import *

class LacssCollaborator(nn.Module):
    """Collaborator module for semi-supervised Lacss training

    Attributes:
        conv_spec: conv-net specificaiton for cell border predicition
        unet_spec: specification for unet, used to predict cell foreground
        patch_size: patch size for the unet
        n_cls: number of classes (cell types) of input images
    """

    conv_spec: Sequence[int] = (32, 32)
    unet_spec: Sequence[int] = (16, 32, 64)
    patch_size: int = 1
    n_cls: int = 1

    @nn.compact
    def __call__(
        self, image: ArrayLike, cls_id: Optional[ArrayLike] = None
    ) -> DataDict:
        assert cls_id is not None or self.n_cls == 1
        c = cls_id.astype(int).squeeze() if cls_id is not None else 0

        net = UNet(self.unet_spec, self.patch_size)
        unet_out = net(image)

        y = unet_out[0]

        fg = nn.Conv(self.n_cls, (3, 3))(y)
        fg = fg[..., c]

        if fg.shape != image.shape[:-1]:
            fg = jax.image.resize(fg, image.shape[:-1], "linear")

        y = image
        for n_features in self.conv_spec:
            y = nn.Conv(n_features, (3, 3), use_bias=False)(y)
            y = nn.GroupNorm(num_groups=None, group_size=1, use_scale=False)(
                y[None, ...]
            )[0]
            y = jax.nn.relu(y)

        y = nn.Conv(self.n_cls, (3, 3))(y)
        cb = y[..., c]

        return dict(
            fg_pred=fg,
            edge_pred=cb,
        )

class _CKSModel(nn.Module):
    principal: Lacss
    collaborator: LacssCollaborator

    def __call__(self, image, cls_id=None, gt_locations=None, *, training=False):
        outputs = self.principal(
            image,
            gt_locations=gt_locations,
            training=training,
        )

        if self.collaborator is not None:
            outputs["predictions"].update(self.collaborator(image=image, cls_id=cls_id))
        return outputs


class LacssTrainer:
    """Main trainer class for Lacss"""

    default_lr: float = 0.001

    def __init__(
        self,
        model_config: dict,
        collaborator_config: Optional[dict] = None,
        *,
        optimizer: Optional[Optimizer] = None,
        seed: Union[int, Array] = 42,
        strategy: type = JIT,
    ):
        """Constructor

        Args:
            config: configuration dictionary for Lacss model
            collaborator_config: configuration dictionary for the collaborator model used in
                weakly-supervised training. If set to None, then no collaborator model will be
                created. In this case, training with weak-supervision will result in a error.

        Keyword Args:
            seed: RNG seed
            optimizer: Override the default optimizer
            strategy: Training backend. See See [Traing backends](/api/train/#training-backends).
        """
        principal = Lacss.from_config(model_config)

        if collaborator_config is None:
            collaborator = None
        else:
            collaborator = LacssCollaborator(**collaborator_config)

        self.model = _CKSModel(
            principal=principal,
            collaborator=collaborator,
        )

        self.optimizer = optimizer or optax.adamw(LacssTrainer.default_lr)
        self.seed = seed
        self.strategy = strategy

    def _train_to_next_interval(self, n_steps, trainer, train_iter, cpm, val_dataset):
        for _ in tqdm(range(n_steps)):
            next(train_iter)

        print("Losses: " + repr(train_iter.loss_logs))

        if cpm is not None:
            cpm.save(
                (cpm.latest_step() or 0) + 1,
                args=ocp.args.StandardSave(train_iter),
            )

        if val_dataset is not None:
            val_metrics = [
                lacss.metrics.LoiAP([5, 2, 1]),
                lacss.metrics.BoxAP([0.5, 0.75]),
            ]

            var_logs = trainer.compute_metrics(
                val_dataset,
                val_metrics,
                dict(params=train_iter.parameters),
            )

            for k, v in var_logs.items():
                print(f"{k}: {v}")

        train_iter.reset_loss_logs()

    def _get_warmup_trainer(self, sigma, pi):
        pre_seg_loss = partial(segmentation_loss, pretraining=True)
        pre_col_seg_loss = partial(collaborator_segm_loss, sigma=100.0, pi=1.0)
        losses = [
            lpn_loss,
            pre_seg_loss,
            pre_col_seg_loss,
            collaborator_border_loss,
            aux_size_loss,
        ]
        return Trainer(
            model=self.model,
            losses=losses,
            optimizer=self.optimizer,
            seed=self.seed,
            strategy=self.strategy,
        )

    def _get_trainer(self, sigma, pi):
        col_seg_loss = partial(collaborator_segm_loss, sigma=sigma, pi=pi)
        losses = [
            lpn_loss,
            segmentation_loss,
            col_seg_loss,
            collaborator_border_loss,
            aux_size_loss,
        ]
        return Trainer(
            model=self.model,
            losses=losses,
            optimizer=self.optimizer,
            seed=self.seed,
            strategy=self.strategy,
        )

    def get_init_params(self, dataset):
        trainer = self._get_trainer(20.0, 2.0)
        train_it = trainer.train(
            dataset, 
            rng_cols=["droppath"],
            training=True,
        )
        return train_it.parameters


    def do_training(
        self,
        dataset: Iterable,
        val_dataset: Iterable | None = None,
        n_steps: int = 50000,
        validation_interval: int = 5000,
        checkpoint_manager: Optional[ocp.CheckpointManager] = None,
        *,
        warmup_steps: int = 0,
        sigma: float = 20.0,
        pi: float = 2.0,
        init_vars = None,
    ) -> None:
        """Runing training.

        Args:
            dataset: An data iterator feed training data. The data should be in the form of a tuple:
                (x, y).
                    x is a dictionary with at least two keys:
                        "image": Trainging image.
                        "gt_locations": Point labels. Nx2 array
                    y is a dictionary of extra labels. It can be
                        None: for point-supervised training
                        "gt_labels": A index-label image (H, W). For segmentation label.
                        "gt_image_mask": A binary image (H, W). For weakly supervised training
            n_steps: Total training steps
            validation_inteval: Step intervals to perform validation and checkpointing.
            val_dataset: If not None, performing validation on this dataset. The data should be in the
                form of a tuple (x, y):
                    x is a dictionary with one key: "image"
                    y is a dictionary with two lables: "gt_bboxes" and "gt_locations"
            checkpoint_manager: If supplied, will be used to created checkpoints. A checkpoint manager
                can be obtained by calling:
                ```
                options = orbax.CheckpointManagerOptions(...)
                manager = orbax.checkpoint.CheckpointManager(
                    'path/to/directory/',
                    options = options
                )
                ```


        keyword Args:
            warmup_steps: Only used for point-supervised training. Pretraining steps, for which a large
                sigma values is used. This should be multiples of validation_inteval
            sigma: Only for point-supervised training. Expected cell size
            pi: Only for point-supervised training. Amplitude of the prior term.
        """
        cur_step = 0

        if cur_step < warmup_steps:
            trainer = self._get_warmup_trainer(sigma, pi)
            train_it = trainer.train(
                dataset, 
                rng_cols=["droppath"], 
                init_vars=init_vars,
                training=True,
            )

            while cur_step < warmup_steps:
                next_cp_step = (
                    (cur_step + validation_interval)
                    // validation_interval
                    * validation_interval
                )

                print(f"Current step {cur_step} going to {next_cp_step}")

                self._train_to_next_interval(
                    next_cp_step - cur_step,
                    trainer,
                    train_it,
                    checkpoint_manager,
                    val_dataset,
                )

                cur_step = next_cp_step

                self.parameters = train_it.parameters

            init_vars = dict(params=train_it.parameters)

        trainer = self._get_trainer(sigma, pi)
        train_it = trainer.train(
            dataset,
            init_vars=init_vars,
            training=True,
        )

        while cur_step < n_steps:
            next_cp_step = (
                (cur_step + validation_interval)
                // validation_interval
                * validation_interval
            )

            print(f"Current step {cur_step} going to {next_cp_step}")

            self._train_to_next_interval(
                next_cp_step - cur_step,
                trainer,
                train_it,
                checkpoint_manager,
                val_dataset,
            )

            cur_step = next_cp_step

            self.parameters = train_it.parameters


    def save(self, save_path) -> None:
        """Save a pickled copy of the Lacss model in the form of (module:Lacss, weights:FrozenDict). Only saves the principal model.

        Args:
            save_path: Path to the pkl file
        """
        import pickle

        _module = self.model.principal
        _params = self.parameters["principal"]

        with open(save_path, "wb") as f:
            pickle.dump((_module, _params), f)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/lacss.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/lacss.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from biopb.lacss import detection_request_pb2 as biopb_dot_lacss_dot_detection__request__pb2
from biopb.lacss import detection_response_pb2 as biopb_dot_lacss_dot_detection__response__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x62iopb/lacss/lacss.proto\x12\x0b\x62iopb.lacss\x1a#biopb/lacss/detection_request.proto\x1a$biopb/lacss/detection_response.proto2\xaf\x01\n\x05Lacss\x12M\n\x0cRunDetection\x12\x1d.biopb.lacss.DetectionRequest\x1a\x1e.biopb.lacss.DetectionResponse\x12W\n\x12RunDetectionStream\x12\x1d.biopb.lacss.DetectionRequest\x1a\x1e.biopb.lacss.DetectionResponse(\x01\x30\x01\x42\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.lacss_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_LACSS']._serialized_start=116
  _globals['_LACSS']._serialized_end=291
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import Optional, Sequence, Callable

import flax.linen as nn
import jax.numpy as jnp

from .convnext import ConvNeXt
from .integrators import FPN
from ..typing import DataDict, Array, ArrayLike, Any
from .common import DefaultUnpicklerMixin
from .video_integrator import VideoIntegrator

class Backbone(nn.Module, DefaultUnpicklerMixin):
    base_type: str = "tiny"
    patch_size: int = 4
    fpn_dim: int = 384
    out_dim: int = 256
    det_layer: int = 0
    seg_layer: int = 0
    n_layers_det: int = 4
    n_layers_seg: int = 3
    drop_path_rate: float=0.4
    activation: Callable[[ArrayLike], Array]=nn.gelu
    deterministic: bool|None = None
    dtype: Any = None

    def setup(self):
        self.cnn = ConvNeXt.get_preconfigured(
            self.base_type, 
            patch_size=self.patch_size, 
            drop_path_rate = self.drop_path_rate, 
            dtype=self.dtype
        )

    def get_ref(self, x):
        if x.ndim == 3:
            x = x[None, ...]
            return self.cnn(x, deterministic=True)

        else:
            assert x.ndim == 4

            to_shape = (-1, self.patch_size) + x.shape[1:]
            image_ = x
            image_ = image_.reshape(to_shape).mean(axis=1)
            f0 = self.cnn(image_, deterministic=True)

            image_ = x.swapaxes(0,1)
            image_ = image_.reshape(to_shape).mean(axis=1)
            f1 = self.cnn(image_, deterministic=True)

            image_ = x.swapaxes(0,2)
            image_ = image_.reshape(to_shape).mean(axis=1)
            f2 = self.cnn(image_, deterministic=True)

            return (f0, f1, f2)

    @nn.compact
    def get_2d_feature(self, image, video_refs, deterministic):
        x = jnp.asarray(image)
        if image.ndim == 3:
            x = x[None, ...]

        assert x.ndim == 4

        x = self.cnn(x, deterministic=deterministic or video_refs is not None)
        self.sow("intermediates", "encoder", x)

        if video_refs is not None:
            x = VideoIntegrator(dtype=self.dtype)(x, video_refs, deterministic=deterministic)

        x = FPN(self.fpn_dim, activation=self.activation, dtype=self.dtype)(x)

        self.sow("intermediates", "decoder", x)

        x_det = x[self.det_layer]
        for _ in range(self.n_layers_det):
            y = nn.Conv(self.out_dim, (3,3), dtype=self.dtype)(x_det)
            y = nn.GroupNorm(dtype=self.dtype)(y)
            x_det = nn.gelu(y)

        x_seg = x[self.seg_layer]
        for _ in range(self.n_layers_seg):
            y = nn.Conv(self.out_dim, (3,3), dtype=self.dtype)(x_seg)
            y = nn.GroupNorm(dtype=self.dtype)(y)
            x_seg = nn.gelu(y)

        if image.ndim == 3:
            x_det = x_det.squeeze(0)
            x_seg = x_seg.squeeze(0)

        return x_det, x_seg


    def __call__(self, image:ArrayLike, video_refs:tuple|None=None, deterministic:bool|None=True):
        image = jnp.asarray(image)
        if deterministic is None:
            deterministic = self.deterministic

        if image.ndim == 3:
            return self.get_2d_feature(image, video_refs, deterministic)

        else:
            def _get_ref(vf, axis):
                if vf is None:
                    return None
                else:
                    return (vf[0][axis], vf[1])

            img_shape = image.shape
            image_ = image
            image_ = image_.reshape((-1, self.patch_size) + img_shape[1:]).mean(axis=1)
            ref_features = _get_ref(video_refs, 0)
            f0 = self.get_2d_feature(image_, ref_features, deterministic)

            image_ = image.swapaxes(0,1)
            image_ = image_.reshape((-1, self.patch_size) + img_shape[1:]).mean(axis=1)
            ref_features = _get_ref(video_refs, 1)
            f1 = self.get_2d_feature(image_, ref_features, deterministic)

            image_ = image.swapaxes(0,2)
            image_ = image_.reshape((-1, self.patch_size) + img_shape[1:]).mean(axis=1)
            ref_features = _get_ref(video_refs, 2)            
            f2 = self.get_2d_feature(image_, ref_features, deterministic)

            x_det = jnp.c_[f0[0], f1[0].swapaxes(0,1), f2[0].swapaxes(0,2)]
            x_seg = jnp.c_[f0[1], f1[1].swapaxes(0,1), f2[1].swapaxes(0,2)]

            return x_det, x_seg
    


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from biopb.lacss import detection_request_pb2 as biopb_dot_lacss_dot_detection__request__pb2
from biopb.lacss import detection_response_pb2 as biopb_dot_lacss_dot_detection__response__pb2

GRPC_GENERATED_VERSION = '1.66.2'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in biopb/lacss/lacss_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class LacssStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RunDetection = channel.unary_unary(
                '/biopb.lacss.Lacss/RunDetection',
                request_serializer=biopb_dot_lacss_dot_detection__request__pb2.DetectionRequest.SerializeToString,
                response_deserializer=biopb_dot_lacss_dot_detection__response__pb2.DetectionResponse.FromString,
                _registered_method=True)
        self.RunDetectionStream = channel.stream_stream(
                '/biopb.lacss.Lacss/RunDetectionStream',
                request_serializer=biopb_dot_lacss_dot_detection__request__pb2.DetectionRequest.SerializeToString,
                response_deserializer=biopb_dot_lacss_dot_detection__response__pb2.DetectionResponse.FromString,
                _registered_method=True)


class LacssServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RunDetection(self, request, context):
        """Unitary call for computing cell detection / segmentation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunDetectionStream(self, request_iterator, context):
        """The streaming version allows dynamic updating of the DetectionRequest and
        getting updated results. For example, a client may request analyses at
        several different settings on the same image. After initial request, the
        following streaming requests no long need to transmit the image pixel
        data anymore, but only the new parameter settings, which saves network
        bandwidth.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LacssServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RunDetection': grpc.unary_unary_rpc_method_handler(
                    servicer.RunDetection,
                    request_deserializer=biopb_dot_lacss_dot_detection__request__pb2.DetectionRequest.FromString,
                    response_serializer=biopb_dot_lacss_dot_detection__response__pb2.DetectionResponse.SerializeToString,
            ),
            'RunDetectionStream': grpc.stream_stream_rpc_method_handler(
                    servicer.RunDetectionStream,
                    request_deserializer=biopb_dot_lacss_dot_detection__request__pb2.DetectionRequest.FromString,
                    response_serializer=biopb_dot_lacss_dot_detection__response__pb2.DetectionResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'biopb.lacss.Lacss', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('biopb.lacss.Lacss', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Lacss(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RunDetection(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/biopb.lacss.Lacss/RunDetection',
            biopb_dot_lacss_dot_detection__request__pb2.DetectionRequest.SerializeToString,
            biopb_dot_lacss_dot_detection__response__pb2.DetectionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RunDetectionStream(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/biopb.lacss.Lacss/RunDetectionStream',
            biopb_dot_lacss_dot_detection__request__pb2.DetectionRequest.SerializeToString,
            biopb_dot_lacss_dot_detection__response__pb2.DetectionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from __future__ import annotations

from typing import Any, Optional, Sequence
from functools import partial
from copy import copy

import jax
import flax.linen as nn
import optax
import ml_collections

from xtrain.utils import unpack_x_y_sample_weight, unpack_prediction_and_state, Inputs
from xtrain.strategy import VMapped, TrainState, LossLog
from xtrain.base_trainer import TrainIterator

from ..utils import deep_update
from ..modules import ConvNeXt
from ..ops import gradient_reversal
from ..typing import ArrayLike, DataDict
from .train import train_fn

from lacss.losses import (
    cks_boundry_loss,
    collaborator_segm_loss,
    self_supervised_instance_loss,
    instance_overlap_loss,
    aux_size_loss,
)

from ..modules import Lacss, UNet
from ..typing import Array, Optimizer

jnp = jax.numpy

class LacssCollaborator(nn.Module):
    """Collaborator module for semi-supervised Lacss training

    Attributes:
        conv_spec: conv-net specificaiton for cell border predicition
        unet_spec: specification for unet, used to predict cell foreground
        patch_size: patch size for the unet
        n_cls: number of classes (cell types) of input images
    """

    conv_spec: Sequence[int] = (32, 32)
    unet_spec: Sequence[int] = (16, 32, 64)

    @nn.compact
    def __call__(self, image: ArrayLike, *args, **kwargs) -> DataDict:
        x = jnp.asarray(image)

        net = UNet(self.unet_spec)
        unet_out = net(x)

        x = unet_out[0]
        foreground = nn.Conv(1, (3, 3))(x).squeeze(-1)

        if foreground.shape != x.shape[:-1]:
            foreground = jax.image.resize(foreground, x.shape[:-1], "linear")

        x = image
        for n_features in self.conv_spec:
            x = nn.Conv(n_features, (3, 3), use_bias=False)(x)
            x = nn.LayerNorm(use_scale=False)(x)
            x = jax.nn.relu(x)

        boundary = nn.Conv(2, (3, 3))(x)

        return dict(
            fg_pred=foreground,
            edge_pred=boundary,
        )


class CKSModel(nn.Module):
    principal: Lacss
    collaborator: LacssCollaborator
    config: ml_collections.ConfigDict

    def __call__(self, **kwargs):
        inputs = Inputs()
        inputs = inputs.update(**kwargs)
        outputs = Inputs.apply(
            train_fn,
            self.principal,
            config = self.config,
        )(inputs)

        outputs["predictions"].update(
            Inputs.apply(self.collaborator)(inputs)
        )

        sigma = self.config.get("sigma", 15.0)
        pi = self.config.get("pi", 2.0)
        w = self.config.get("w", 1e-3)

        outputs['losses']['cks_loss'] = (
            collaborator_segm_loss(kwargs, outputs, sigma=sigma, pi=pi) + 
            cks_boundry_loss(kwargs, outputs) +
            self_supervised_instance_loss(kwargs, outputs) + 
            instance_overlap_loss(kwargs, outputs, soft_label=True) +
            aux_size_loss(kwargs, outputs, weight=w)
        )

        return outputs

def _init_cks(it, model, inputs):
    inputs = Inputs.from_value(inputs)
    rngs = it.rngs.copy()
    rngs['params'] = jax.random.PRNGKey(123)
    cks_params = Inputs.apply(model.collaborator.init, rngs)(inputs)['params']
    params = dict(
        principal = it.parameters,
        collaborator = cks_params,
    )
    state = TrainState.create(
        apply_fn=model.apply,
        params=params,
        tx=it.ctx.optimizer,
    )
    it_ = copy(it)
    it_.train_state = state
    it_.loss_logs = [
        LossLog("losses/lpn_detection_loss"), 
        LossLog("losses/lpn_localization_loss"),
        LossLog("losses/cks_loss")
    ]
    it_.variables = {}
    it_.has_aux = False
    # it_.frozen = jax.tree_util.tree_map(lambda _:False, it_.parameters)
    # it_.freeze("principal")
    return it_


class CKS(VMapped):
    var_key:str = "cks"
    config: ml_collections.ConfigDict = ml_collections.ConfigDict()

    @classmethod
    def train_step(
        cls,
        train_obj: TrainIterator,
        batch: Any,
    ) -> tuple[Any, TrainIterator]:
        inputs, label, _ = unpack_x_y_sample_weight(batch)

        if not cls.var_key in train_obj.variables:
            model = CKSModel(
                train_obj.ctx.model,
                LacssCollaborator(),
                cls.config,
            )
            train_obj.variables[cls.var_key] = _init_cks(train_obj, model, inputs)

        cks = train_obj.variables.pop(cls.var_key)

        if label is None:

            cks.parameters['principal'] = train_obj.parameters
            cks_grad, (prediction, cks) = jax.grad(cls.loss_fn, has_aux=True)(
                cks.parameters, cks, batch
            )
            cks.train_state = cks.train_state.apply_gradients(grads=cks_grad)
            del cks.train_state.params['principal']

            grads = cks_grad['principal']

        else:
            grads, (prediction, train_obj) = jax.grad(cls.loss_fn, has_aux=True)(
                train_obj.parameters, train_obj, batch,
            )

        train_obj.train_state = train_obj.train_state.apply_gradients(grads=grads)

        train_obj.variables[cls.var_key] = cks

        return prediction, train_obj



from __future__ import annotations

from functools import partial
import jax
import jax.numpy as jnp

from ..typing import *
from .boxes import box_iou_similarity, distance_similarity

NMS_TILE_SIZE = 1024

def _suppress(boxes, mask):
    return jnp.where(mask[:, None], -1, boxes)


def _suppression_loop_body(inputs, *, similarity_func=None):
    """Process boxes in the range [idx*NMS_TILE_SIZE, (idx+1)*NMS_TILE_SIZE).
    Args:
        inputs: tuple
            idx: current slice
            boxes: represetn bboxes if shape is [N, 4] else represent points [N, 2/3].
            num_selected: number of selected boxes so far
            threshold: float
    Returns:
        boxes: updated
        num_selected: updated
    """

    # Iterates over tiles that can possibly suppress the current tile.

    idx, boxes, num_selected, threshold = inputs

    box_slice = boxes[idx]

    if similarity_func is None:
        similarity_func = (
            box_iou_similarity if boxes.shape[-1] > 3 else distance_similarity
        )

    def _for_loop_func(idx, slice):
        iou = similarity_func(boxes[idx], slice)
        suppressed = jnp.any(iou >= threshold, axis=0)
        return _suppress(slice, suppressed)

    box_slice = jax.lax.fori_loop(0, idx, _for_loop_func, box_slice)

    # Iterates over the current tile to compute self-suppression.
    iou = similarity_func(box_slice, box_slice)
    mask = jnp.arange(NMS_TILE_SIZE).reshape([1, -1]) > jnp.arange(
        NMS_TILE_SIZE
    ).reshape([-1, 1])
    mask = mask & (iou > threshold)

    def _while_loop_func(inputs):
        mask, cnt = inputs
        cnt = jnp.count_nonzero(mask)
        can_suppress_others = ~mask.any(axis=0)
        suppressed = (mask & can_suppress_others[:, None]).any(axis=0)
        mask = mask & ~suppressed[:, None]
        # suppressed = mask.at[can_suppress_others, :].any(axis=0)
        # mask = mask.at[suppressed, :].set(False)
        return mask, cnt

    def _while_cond_func(inputs):
        mask, cnt = inputs
        return jnp.count_nonzero(mask) != cnt

    mask, _ = jax.lax.while_loop(_while_cond_func, _while_loop_func, (mask, 0))
    box_slice = _suppress(box_slice, mask.any(axis=0))

    # Uses box_slice to update the input boxes.
    boxes = boxes.at[idx].set(box_slice)

    # output_size.
    num_selected += jnp.count_nonzero((box_slice >= 0).any(axis=-1))

    return boxes, num_selected


def _nms(
    scores: ArrayLike,
    boxes: ArrayLike,
    max_output_size: int,
    threshold: float = 0.5,
    min_score: float = 0,
    similarity_func: callable|None = None,
) -> Array:
    # preprocessing
    c = boxes.shape[-1]

    # if c != 2 and c != 4:
        # raise ValueError(f"boxes should be Nx4 or Nx2, got Nx{c}")

    num_boxes = boxes.shape[0]
    pad = NMS_TILE_SIZE - 1 - (num_boxes - 1) % NMS_TILE_SIZE
    boxes = jnp.pad(boxes, [[0, pad], [0, 0]], constant_values=-1)
    scores = jnp.pad(scores, [[0, pad]], constant_values=-1)
    orig_num_boxes = num_boxes
    num_boxes += pad
    boxes = boxes.reshape(-1, NMS_TILE_SIZE, c)

    # process all tiles until generating enough output
    def _trivial_suppress_all(inputs):
        (
            idx,
            boxes,
            num_outputs,
            _,
        ) = inputs
        boxes = boxes.at[idx].set(-1)
        return boxes, num_outputs

    def _inner_loop_func(idx, val):
        boxes, num_selected = val
        return jax.lax.cond(
            (scores[idx * NMS_TILE_SIZE] >= min_score) & (num_selected < max_output_size),
            partial(_suppression_loop_body, similarity_func=similarity_func),
            _trivial_suppress_all,
            (idx, boxes, num_selected, threshold),
        )

    num_selected = 0
    boxes, num_selected = jax.lax.fori_loop(
        0,
        num_boxes // NMS_TILE_SIZE,
        _inner_loop_func,
        (boxes, num_selected),
    )

    # reshape boxes back
    boxes = boxes.reshape(-1, c)

    # find valid boxes
    selected = (boxes >= 0).any(axis=-1)
    selected &= scores >= min_score

    # remove padding
    selected = selected[:orig_num_boxes]

    return selected


def non_max_suppression(
    scores: ArrayLike,
    boxes: ArrayLike,
    max_output_size: int,
    threshold: float = 0.5,
    min_score: float = 0,
    return_selection: bool = False,
    similarity_func: callable|None = None,
) -> Array|tuple[Array]:
    """non-maximum suppression for either bboxes or points.

    Assumption:

        * The boxes/points are sorted by scores

    The overal design of the algorithm is to handle boxes tile-by-tile:

    Args:
        scores: [N]
        boxes: [N, C]  C=4 for boxes, C=2/3 for locations
        max_output_size: a positive scalar integer
        threshold: threshold of similarity score to supress
        min_score: min score to be selected, default 0
        return_selection: whether also return the boolean indicator
        similarity_func: Optionally provide a custom callable to compute similarity score

    Returns:
        nms_scores: [M].  M = max_output_size
        nms_proposals: [M, C].
        selection: [N] a boolean indicator of selection status of original input
          only if return_selection is True
    """
    if max_output_size <= 0:
        max_output_size = boxes.shape[0]

    selected = _nms(
        scores,
        boxes,
        max_output_size,
        threshold,
        min_score,
    )

    if return_selection:
        return selected
    else:
        # find index of selected
        idx_of_selected = jnp.argwhere(
            selected, size=max_output_size, fill_value=-1
        ).squeeze(-1)

        scores = jnp.where(idx_of_selected >= 0, scores[idx_of_selected], -1.0)
        boxes = jnp.where(idx_of_selected[:, None] >= 0, boxes[idx_of_selected], -1.0) 
               
        return scores, boxes



from __future__ import annotations

import numpy as np

class Dice:
    """Compute instance Dice values"""

    def __init__(self):
        self.pred_areas = []
        self.gt_areas = []
        self.pred_scores = []
        self.gt_scores = []

    def update(self, pred_its, pred_areas, gt_areas):
        self.pred_areas.append(pred_areas)
        self.gt_areas.append(gt_areas)

        pred_best = pred_its.max(axis=1)
        pred_best_matches = pred_its.argmax(axis=1)
        pred_dice = pred_best * 2 / (pred_areas + gt_areas[pred_best_matches])
        self.pred_scores.append(pred_dice)

        gt_best = pred_its.max(axis=0)
        gt_best_matches = pred_its.argmax(axis=0)
        gt_dice = gt_best * 2 / (gt_areas + pred_areas[gt_best_matches])
        self.gt_scores.append(gt_dice)

    def compute(self):
        pred_areas = np.concatenate(self.pred_areas)
        gt_areas = np.concatenate(self.gt_areas)
        pred_scores = np.concatenate(self.pred_scores)
        gt_scores = np.concatenate(self.gt_scores)

        pred_dice = (pred_areas / pred_areas.sum() * pred_scores).sum()
        gt_dice = (gt_areas / gt_areas.sum() * gt_scores).sum()

        dice = (pred_dice + gt_dice) / 2

        return dice




from __future__ import annotations

import jax
import optax
from ml_collections import ConfigDict

from ..losses import binary_focal_crossentropy
from ..ops import match_and_replace, distance_similarity
from ..typing import ArrayLike
from ..utils import deep_update

def train_fn(
    module,
    image: ArrayLike,
    gt_locations: ArrayLike,
    image_mask: ArrayLike|None = None,
    config: ConfigDict = ConfigDict(),
):
    pass



from typing import Sequence

from dataclasses import dataclass
import logging

import numpy as np
import scipy

from .kalman import KalmanFilter, KTracker
from ..ops import iou_loss, box_iou_similarity, generalized_iou_loss

def _from_obs(yxhw):
    return np.r_[yxhw[:2] - yxhw[2:]/2 , yxhw[:2] + yxhw[2:]/2]

def _to_obs(yxyx):
    return np.r_[(yxyx[:2] + yxyx[2:])/2 , (yxyx[2:] - yxyx[:2])]

def _nms_primary(boxes, scores, score_threshold, iou_threshold):
    ious = box_iou_similarity(boxes, boxes)
    selected = scores > score_threshold
    for irow in range(scores.shape[0]):
        if selected[irow]:
            supress = ious[irow] > iou_threshold
            supress[:(irow + 1)] = False
            selected[supress] = False
    return selected

def _nms_secondary(selected_boxes, boxes, iou_threshold):
    ious = box_iou_similarity(selected_boxes, boxes)
    supress = (ious > iou_threshold).any(axis=0)
    selected = ~ supress
    ious = box_iou_similarity(boxes, boxes)
    for irow in range(ious.shape[0]):
        if selected[irow]:
            supress = ious[irow] > iou_threshold
            supress[:(irow + 1)] = False
            selected[supress] = False
    return selected


@dataclass
class ZStackTracker:
    """ A special tracker for building 3d segmentation from a z-stack of 2d segmentations. It uses a kalman
    tracker of 2d boxes that assumes a fixed center location and tracking only the size changes.

    Attributes:
        score_thresh: score threshold for high-confidence detections
        nms_iou: filter the 2D results with non-max-supression
        cost_thresh: cost threshold for link assignment
        cost_thresh_secondary:  cost threshold for link assignment of low confidence detections
        max_init_area: maximum segmentation area to be considered as the first slice of a cell 
        max_time_lost: max consecutive missing slices  
        std_weight_position: kalman filter parameter. relative error for position 
        std_weight_velocity: kalman filter parameter. relative error for velocity
        min_z_slices: minimal z slices needed to be considered a cell
        use_generalized_iou_loss: whether to use generalized iou loss instead of simple iou_loss for linking
            cost
    """
    score_thresh: float = 0.5 # threshold for high-confidence detections
    nms_iou: float = 0.3
    cost_thresh: float = 0.65 # cost threshold for link assignment
    cost_thresh_secondary: float = 0.5 # cost threshold for link assignment of low confidence detections
    max_init_area: float = 1000
    max_time_lost: int = 1
    std_weight_position: float = 1/20
    std_weight_velocity: float = 1/160
    min_z_slices: int = 3
    use_generalized_iou_loss: bool = False

    # tracks y, x, h, w, vh, vw
    class ZStack_KF(KalmanFilter):
        def __init__(self, std_weight_position, std_weight_velocity):
            self._motion_mat = np.eye(6, 6)
            self._motion_mat[2, 4] = 1
            self._motion_mat[3, 5] = 1
            self._update_mat = np.eye(4, 6)
            self._std_weight_position = std_weight_position
            self._std_weight_velocity = std_weight_velocity

        def initiate(self, measurement):
            mean_pos = measurement
            mean_vel = np.zeros([2])
            mean = np.r_[mean_pos, mean_vel]
            std = [
                2 * self._std_weight_position * measurement[2],
                2 * self._std_weight_position * measurement[3],
                2 * self._std_weight_position * measurement[2],
                2 * self._std_weight_position * measurement[3],
                10 * self._std_weight_velocity * measurement[2],
                10 * self._std_weight_velocity * measurement[3],
            ]
            covariance = np.diag(np.square(std))

            return mean, covariance

        def predict(self, mean, covariance):
            mean = np.dot(mean, self._motion_mat.T)

            std = [
                0, # self._std_weight_position * mean[2],
                0, # self._std_weight_position * mean[3],
                self._std_weight_position * mean[2],
                self._std_weight_position * mean[3],
                self._std_weight_velocity * mean[2],
                self._std_weight_velocity * mean[3],
            ]
            covariance = np.linalg.multi_dot((
                self._motion_mat, covariance, self._motion_mat.T)
            ) + np.diag(np.square(std))

            return mean, covariance

        def project(self, mean, covariance):
            std = [
                0, #self._std_weight_position * mean[2],
                0, #self._std_weight_position * mean[3],
                self._std_weight_position * mean[2],
                self._std_weight_position * mean[3],
            ]
            innovation_cov = np.diag(np.square(std))

            mean = np.dot(self._update_mat, mean)
            covariance = np.linalg.multi_dot((
                self._update_mat, covariance, self._update_mat.T
            ))
            return mean, covariance + innovation_cov

        def update(self, mean, covariance, measurement):
            projected_mean, projected_cov = self.project(mean, covariance)

            chol_factor, lower = scipy.linalg.cho_factor(
                projected_cov, lower=True, check_finite=False)
            kalman_gain = scipy.linalg.cho_solve(
                (chol_factor, lower), np.dot(covariance, self._update_mat.T).T,
                check_finite=False).T
            innovation = measurement - projected_mean

            new_mean = mean + np.dot(innovation, kalman_gain.T)
            new_covariance = covariance - np.linalg.multi_dot((
                kalman_gain, projected_cov, kalman_gain.T))

            new_mean[:2] = (mean[:2] + new_mean[:2]) / 2
            new_covariance[:2, :2] = (new_covariance[:2, :2] + projected_cov[:2, :2])/4

            return new_mean, new_covariance

    def __post_init__(self):
        self._kf = self.ZStack_KF(std_weight_position=self.std_weight_position, std_weight_velocity=self.std_weight_velocity)


    def _assign(self, tracks, dets, threshold):
        box_a = np.asarray([_from_obs(t.mean[:4]) for t in tracks]).reshape(-1, 4)
        box_b = np.asarray([_from_obs(t.obs) for t in dets]).reshape(-1, 4)

        if self.use_generalized_iou_loss:
            cost_matrix = generalized_iou_loss(box_a, box_b)
        else:   
            cost_matrix = iou_loss(box_a, box_b)

        return KTracker.assign(tracks, dets, cost_matrix, threshold)

    def update(self, tracks:Sequence[KTracker], dets:dict, frame_id: int) -> tuple[list[KTracker], list[KTracker]]:
        """ Update the tracker with one frame

        Args:
            tracks: a list of KTacker representing currently tracks cells
            dets: predictor output in bbox format
            frame_id: current frame number

        Returns:
            tracked_tracks: list of tracks that are active
            removed_tracks: list of tracks that are inactive

        """
        scores = np.asarray(dets["pred_scores"]).reshape(-1)
        bboxes = np.asarray(dets["pred_bboxes"]).reshape(-1, 4)
        assert len(scores) == bboxes.shape[0]
        assert bboxes.shape[1] == 4
        flatten_dets = np.asarray([dict(zip(dets.keys(), x)) for x in zip(*dets.values())], dtype=object)
        detections = np.asarray([ 
            KTracker(
                _to_obs(box),
                frame_id,
                self._kf,
                data = data,
            ) for box, data in zip(bboxes, flatten_dets)
        ], dtype=object)

        logging.info(f"processing frame {frame_id}: {len(tracks)} input tracks and {len(flatten_dets)} new detections.")

        inds_primary = _nms_primary(bboxes, scores, self.score_thresh, self.nms_iou)
        inds_secondary = ~ inds_primary

        # find current pool of tracks
        removed_tracks, track_pool = [], []
        for track in tracks:
            if track.state == "tracked" or track.state == "lost":
                track.predict()
                track_pool.append(track)
            else:
                raise ValueError(f"Invalid track state {track.state}.")

        ''' First association, with high score detection boxes'''
        tracked_tracks, track_pool, detections_primary = self._assign(
            track_pool, 
            detections[inds_primary],
            threshold=self.cost_thresh,
        )

        ''' Second association, with low score detection boxes'''
        lost_tracks, r_track_pool = [], []
        for track in track_pool:
            if track.state == "tracked":
                r_track_pool.append(track)
            else:
                lost_tracks.append(track)

        cur_boxes = bboxes[inds_primary]
        inds_secondary = _nms_secondary(cur_boxes, bboxes, self.nms_iou)

        matched, unmatched, detections_secondary = self._assign(
            r_track_pool, 
            detections[inds_secondary], 
            threshold=self.cost_thresh_secondary,
        )
        tracked_tracks += matched
        lost_tracks += unmatched

        n_tracks = len(tracked_tracks)
        logging.info(f"processing frame {frame_id}: found {n_tracks} connections")

        """ remove tracks that have been lost for too long """
        for track in lost_tracks:
            if frame_id - track.frame_id > self.max_time_lost:
                lost_tracks.remove(track)
                removed_tracks.append(track)

        logging.info(f"processing frame {frame_id}: remove {len(removed_tracks)} tracks that has lost for too long")

        """ Init new stracks"""
        cur_boxes = np.asarray([t.history[-1]["pred_bboxes"] for t in tracked_tracks]).reshape(-1, 4)
        inds_remaining = _nms_secondary(cur_boxes, bboxes, self.nms_iou)
        for track in detections[inds_remaining]:
            if track._data["pred_scores"] >= 0.5:
                if track.obs[2:].prod() < self.max_init_area:
                    track.initialize()
                    tracked_tracks.append(track)

        logging.info(f"processing frame {frame_id}: initialize {len(tracked_tracks) - n_tracks} new detections")

        # sanity check
        cur_boxes = np.asarray([t.history[-1]["pred_bboxes"] for t in tracked_tracks]).reshape(-1, 4)
        ious = box_iou_similarity(cur_boxes, cur_boxes)
        ious[(range(len(cur_boxes)), range(len(cur_boxes)))] = 0
        assert not ( ious > self.nms_iou ).any()

        """ Update state"""
        for t in lost_tracks:
            m = t.mean
            m[-2:] = 0 # clear velocity for lost tracks
            t.mark_lost(update_state=m)

        return tracked_tracks + lost_tracks, removed_tracks

    def finalize(self, tracks: list[KTracker], *, fill_in_missing:bool=True) -> list[KTracker]:
        """ Finalize step which compute scores and bboxes, and optionally fill in the missing segmentations.
        The aggregated score is the mean of the top-k 2d scores, where k is the min_z_slices. 

        Args:
            tracks: list of tracks, each represent a cell in 3d
        
        Keyword Args:
            fill_in_missing: whether try to generate segmentations for missing slices

        Returns:
            list of tracks will "score", "bbox" fields.        
        """
        tracks = [track for track in tracks if track.frame_id - track.start_frame + 1 >= self.min_z_slices]

        if fill_in_missing:
            for track in tracks:
                data0 = track.history[0]
                f0 = data0["frame_id"] - 1
                for idx, data in enumerate(track.history):
                    if data["frame_id"] != f0 + 1: # missing frame
                        new_data = data.copy()
                        f0 = f0 + 1
                        new_data.update(dict(
                            frame_id = f0,
                            pred_scores = -1,
                            pred_bboxes = (data0["pred_bboxes"] + data["pred_bboxes"])/2,
                            pred_masks = (data["pred_masks"] + data0["pred_masks"]) / 2
                        ))
                        track.history.insert(idx, new_data)
                    else:
                        f0 = data["frame_id"]
                        data0 = data
        
        for cell_data in tracks:
            z0 = cell_data.start_frame
            z1 = cell_data.frame_id + 1
            scores, bboxes = [], []
            for slice_data in cell_data.history:
                scores.append(max(slice_data["pred_scores"], 0.))
                bboxes.append(slice_data["pred_bboxes"])
            bboxes = np.stack(bboxes)
            scores = np.stack(scores)
            top_score_idx = np.argpartition(scores, -self.min_z_slices)[-self.min_z_slices:]
            cell_data.score = scores[top_score_idx].mean()
            cell_data.bbox = np.r_[z0, bboxes[:, :2].min(axis=0), z1, bboxes[:, 2:].max(axis=0)]            

        return tracks

    @staticmethod
    def render_label(tracks:list[KTracker], img3d_shape:Sequence[int])->np.ndarray:
        """ create 3d label from tracking results

        Args:
            tracks: list of tracks
            img3d_shape: tuple of (D, H, W)

        Returns:
            label: array of shape (D, H, W) 
        """
        from skimage.transform import resize
        label = np.zeros(img3d_shape, dtype="uint8")
        _, h, w = img3d_shape

        score_idx = np.argsort([t.score for t in tracks])

        for c, track_id in enumerate(score_idx):
            track = tracks[track_id]
            for data in track.history:
                box = np.round(data["pred_bboxes"]).astype(int)
                box = np.minimum([h,w,h,w], box)
                box = np.maximum(0, box)
                mask = resize(data["pred_masks"], box[2:] - box[:2])
                mask = (mask >= 0.5) * (c+1)
                f = data["frame_id"]
                label[f, box[0]:box[2], box[1]:box[3]] = np.maximum(mask, label[f, box[0]:box[2], box[1]:box[3]])
        
        return label



from __future__ import annotations

from dataclasses import field

import jax
import flax.linen as nn

from .backbone import Backbone
from .lpn import LPN
from .segmentor import Segmentor
from .lpn3d import LPN3D
from .segmentor3d import Segmentor3D
from .common import DefaultUnpicklerMixin

from ..utils import deep_update
from ..typing import ArrayLike, Array, Any

jnp = jax.numpy

class Lacss(nn.Module, DefaultUnpicklerMixin):
    """Main class for LACSS model

    Attributes:
        backbone: CNN backbone
        detector: detection head to predict cell locations
        segmentor: The segmentation head
    """
    backbone: nn.Module = field(default_factory=Backbone)
    detector: nn.Module = field(default_factory=LPN)
    segmentor: nn.Module | None = field(default_factory=Segmentor)

    detector_3d: nn.Module | None = None
    segmentor_3d: nn.Module | None = None

    def get_features(self, image, ref_features, deterministic):
        return self.backbone(image, ref_features, deterministic=deterministic)

    def __call__(self, image: ArrayLike, *, image_mask:ArrayLike|None=None, video_refs:tuple|None=None)->dict:
        """
        Args:
            image: [H, W, C] or [D, H, W, C]
        Returns:
            a dict of model outputs

        """
        # add a batch dim if needed
        image = jnp.asarray(image)

        x_det, x_seg = self.get_features(image, video_refs, deterministic=True)
        outputs = dict(
            det_features = x_det,
            seg_features = x_seg,
        )

        if image.ndim == 3:
            detector_outputs = self.detector(x_det, mask=image_mask)
            outputs = deep_update(outputs, detector_outputs)

            if self.segmentor is not None:
                segmentor_out = self.segmentor(
                    x_seg, outputs["predictions"]["locations"],
                )
                outputs = deep_update(outputs, segmentor_out)

        else:
            if self.detector_3d is None:
                raise(ValueError("not a 3d detector"))
            detector_outputs = self.detector_3d(x_det, mask=image_mask)
            outputs = deep_update(outputs, detector_outputs)

            if self.segmentor_3d is not None:
                segmentor_out = self.segmentor_3d(
                    x_seg, outputs["predictions"]["locations"],
                )
                outputs = deep_update(outputs, segmentor_out)            

        return outputs


    def get_config(self):
        from dataclasses import asdict
        from ..utils import remove_dictkey

        cfg = asdict(self)

        remove_dictkey(cfg, "parent")

        return cfg


    @classmethod
    def from_config(cls, config):
        def _build_sub_module(module, k):
            if k in config and config[k] is not None:
                return module(**config[k])
            else:
                return None

        if isinstance(config, str):
            return cls.get_preconfigued(config)

        else:
            backbone = _build_sub_module(Backbone, "backbone")
            detector = _build_sub_module(LPN, "detector")

            assert backbone is not None and detector is not None, f"invalid config"

            obj = Lacss(
                backbone=backbone,
                detector=detector,
                segmentor=_build_sub_module(Segmentor, "segmentor"),
                detector_3d = _build_sub_module(LPN3D, "detector_3d"),
                segmentor_3d = _build_sub_module(Segmentor3D, "segmentor_3d"),
            )

            return obj

    @classmethod
    def get_default_model(cls, dtype=None):
        return cls.get_base_model(dtype=dtype)

    @classmethod
    def get_small_model(cls, dtype=None):
        return cls(
            backbone=Backbone(
                "tiny", 
                fpn_dim = 384,
                out_dim = 256,
                dtype=dtype,
            ),
            detector=LPN(dtype=dtype),
            segmentor=Segmentor(
                patch_dim = 32,
                sig_dim = 512,
                pos_emb_shape=(16,16,4), 
                dtype=dtype,
            ),
            detector_3d=LPN3D(dim=384, dtype=dtype),
            segmentor_3d=Segmentor3D(
                patch_dim=48, 
                sig_dim=1024, 
                pos_emb_shape=(8,8,8,4),
                dtype=dtype,
            ),
        )

    @classmethod
    def get_base_model(cls, dtype=None):
        return cls(
            backbone=Backbone(
                "base_v2",
                fpn_dim = 512, 
                out_dim = 384, 
                dtype=dtype
            ),
            detector=LPN(dtype=dtype),
            segmentor=Segmentor(
                patch_dim=64, 
                sig_dim=768, 
                pos_emb_shape=(16,16,6), 
                dtype=dtype,
            ),
            detector_3d=LPN3D(dim=512, dtype=dtype),
            segmentor_3d=Segmentor3D(
                patch_dim=64, 
                sig_dim=1384, 
                pos_emb_shape=(8,8,8,6),
                dtype=dtype,
            ),
        )


    @classmethod
    def get_large_model(cls, dtype=None):
        return cls(
            backbone=Backbone(
                "large_v2", 
                fpn_dim = 768,
                out_dim = 512, 
                dtype=dtype
            ),
            detector=LPN(dtype=dtype),
            segmentor=Segmentor(
                patch_dim=96,
                sig_dim=1024,
                pos_emb_shape=(16,16,8),
                dtype=dtype,
            ),
            detector_3d=LPN3D(dim=768, dtype=dtype),
            segmentor_3d=Segmentor3D(
                patch_dim=96, 
                sig_dim=2048, 
                pos_emb_shape=(8,8,8,8),
                dtype=dtype,
            ),
        )

    @classmethod
    def get_preconfigued(cls, config: str, dtype=None):
        if config == "default" or config == "base":
            return cls.get_default_model(dtype=dtype)
        elif config == "small":
            return cls.get_small_model(dtype=dtype)
        elif config == "large":
            return cls.get_large_model(dtype=dtype)
        else:
            raise ValueError(f"Unkown model config {config}")



import sys
from pathlib import Path
sys.path.append(str(Path(__file__).parent)) # hack to get around grpc_python's regid design

from biopb.lacss.lacss_pb2_grpc import Lacss, LacssServicer, LacssStub, add_LacssServicer_to_server
from biopb.lacss.detection_request_pb2 import DetectionRequest
from biopb.lacss.detection_response_pb2 import DetectionResponse, ScoredROI
from biopb.lacss.bindata_pb2 import BinData
from biopb.lacss.detection_settings_pb2 import DetectionSettings
from biopb.lacss.image_data_pb2 import ImageData, Pixels, ImageAnnotation
from biopb.lacss.roi_pb2 import ROI, Rectangle, Mask, Mesh, Polygon, Point



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/detection_settings.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/detection_settings.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$biopb/lacss/detection_settings.proto\x12\x0b\x62iopb.lacss\"\x8e\x02\n\x11\x44\x65tectionSettings\x12\x1a\n\rmin_cell_area\x18\x01 \x01(\x02H\x01\x88\x01\x01\x12\x16\n\tmin_score\x18\x02 \x01(\x02H\x02\x88\x01\x01\x12#\n\x16segmentation_threshold\x18\x03 \x01(\x02H\x03\x88\x01\x01\x12\x14\n\x07nms_iou\x18\x04 \x01(\x02H\x04\x88\x01\x01\x12\x1c\n\x12\x63\x65ll_diameter_hint\x18\x06 \x01(\x02H\x00\x12\x16\n\x0cscaling_hint\x18\x07 \x01(\x02H\x00\x42\r\n\x0bresize_infoB\x10\n\x0e_min_cell_areaB\x0c\n\n_min_scoreB\x19\n\x17_segmentation_thresholdB\n\n\x08_nms_iouB\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.detection_settings_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_DETECTIONSETTINGS']._serialized_start=54
  _globals['_DETECTIONSETTINGS']._serialized_end=324
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/bindata.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/bindata.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x62iopb/lacss/bindata.proto\x12\x0b\x62iopb.lacss\"o\n\x07\x42inData\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x33\n\nendianness\x18\x02 \x01(\x0e\x32\x1f.biopb.lacss.BinData.Endianness\"!\n\nEndianness\x12\x07\n\x03\x42IG\x10\x00\x12\n\n\x06LITTLE\x10\x01\x42\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.bindata_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_BINDATA']._serialized_start=42
  _globals['_BINDATA']._serialized_end=153
  _globals['_BINDATA_ENDIANNESS']._serialized_start=120
  _globals['_BINDATA_ENDIANNESS']._serialized_end=153
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/detection_response.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/detection_response.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from biopb.lacss import roi_pb2 as biopb_dot_lacss_dot_roi__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$biopb/lacss/detection_response.proto\x12\x0b\x62iopb.lacss\x1a\x15\x62iopb/lacss/roi.proto\"9\n\tScoredROI\x12\x1d\n\x03roi\x18\x01 \x01(\x0b\x32\x10.biopb.lacss.ROI\x12\r\n\x05score\x18\x02 \x01(\x02\"?\n\x11\x44\x65tectionResponse\x12*\n\ndetections\x18\x01 \x03(\x0b\x32\x16.biopb.lacss.ScoredROIB\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.detection_response_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_SCOREDROI']._serialized_start=76
  _globals['_SCOREDROI']._serialized_end=133
  _globals['_DETECTIONRESPONSE']._serialized_start=135
  _globals['_DETECTIONRESPONSE']._serialized_end=198
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/roi.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/roi.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from biopb.lacss import bindata_pb2 as biopb_dot_lacss_dot_bindata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x62iopb/lacss/roi.proto\x12\x0b\x62iopb.lacss\x1a\x19\x62iopb/lacss/bindata.proto\"3\n\x05Point\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\x0e\n\x01z\x18\x03 \x01(\x02H\x00\x88\x01\x01\x42\x04\n\x02_z\"[\n\tRectangle\x12$\n\x08top_left\x18\x01 \x01(\x0b\x32\x12.biopb.lacss.Point\x12(\n\x0c\x62ottom_right\x18\x02 \x01(\x0b\x32\x12.biopb.lacss.Point\"Y\n\x04Mask\x12)\n\trectangle\x18\x01 \x01(\x0b\x32\x16.biopb.lacss.Rectangle\x12&\n\x08\x62in_data\x18\x02 \x01(\x0b\x32\x14.biopb.lacss.BinData\"Q\n\x07\x45llipse\x12\"\n\x06\x63\x65nter\x18\x01 \x01(\x0b\x32\x12.biopb.lacss.Point\x12\"\n\x06radius\x18\x02 \x01(\x0b\x32\x12.biopb.lacss.Point\"-\n\x07Polygon\x12\"\n\x06points\x18\x01 \x03(\x0b\x32\x12.biopb.lacss.Point\"\xb1\x01\n\x04Mesh\x12!\n\x05verts\x18\x01 \x03(\x0b\x32\x12.biopb.lacss.Point\x12%\n\x05\x66\x61\x63\x65s\x18\x02 \x03(\x0b\x32\x16.biopb.lacss.Mesh.Face\x12#\n\x07normals\x18\x03 \x03(\x0b\x32\x12.biopb.lacss.Point\x12\x0e\n\x06values\x18\x04 \x03(\x02\x1a*\n\x04\x46\x61\x63\x65\x12\n\n\x02p1\x18\x01 \x01(\r\x12\n\n\x02p2\x18\x02 \x01(\r\x12\n\n\x02p3\x18\x03 \x01(\r\"\xf8\x01\n\x03ROI\x12#\n\x05point\x18\x01 \x01(\x0b\x32\x12.biopb.lacss.PointH\x00\x12+\n\trectangle\x18\x02 \x01(\x0b\x32\x16.biopb.lacss.RectangleH\x00\x12\'\n\x07\x65llipse\x18\x03 \x01(\x0b\x32\x14.biopb.lacss.EllipseH\x00\x12\'\n\x07polygon\x18\x04 \x01(\x0b\x32\x14.biopb.lacss.PolygonH\x00\x12!\n\x04mesh\x18\x05 \x01(\x0b\x32\x11.biopb.lacss.MeshH\x00\x12!\n\x04mask\x18\x06 \x01(\x0b\x32\x11.biopb.lacss.MaskH\x00\x42\x07\n\x05shapeB\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.roi_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_POINT']._serialized_start=65
  _globals['_POINT']._serialized_end=116
  _globals['_RECTANGLE']._serialized_start=118
  _globals['_RECTANGLE']._serialized_end=209
  _globals['_MASK']._serialized_start=211
  _globals['_MASK']._serialized_end=300
  _globals['_ELLIPSE']._serialized_start=302
  _globals['_ELLIPSE']._serialized_end=383
  _globals['_POLYGON']._serialized_start=385
  _globals['_POLYGON']._serialized_end=430
  _globals['_MESH']._serialized_start=433
  _globals['_MESH']._serialized_end=610
  _globals['_MESH_FACE']._serialized_start=568
  _globals['_MESH_FACE']._serialized_end=610
  _globals['_ROI']._serialized_start=613
  _globals['_ROI']._serialized_end=861
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..typing import Array, ArrayLike

class UNet(nn.Module):
    model_spec: Sequence[int] = (32, 64, 128, 256, 512)
    patch_size: int = 1
    se_ratio: int = -1
    is_stack: bool = False

    def __post_init__(self):
        super().__post_init__()
        if not self.patch_size in [1, 2, 4]:
            raise ValueError("patch_size must be eith 1, 2 or 4")

    def _stack_conv(self, x):
        if self.is_stack:
            x = nn.Conv(x.shape[-1], (3, 1, 1), use_bias=False)(x)
        return x

    @nn.compact
    def __call__(self, x: ArrayLike) -> Sequence[Array]:

        encoder_out = []
        fs = max(self.patch_size, 3)
        st = self.patch_size

        ra = tuple(range(x.ndim))

        for ch in self.model_spec:

            x = self._stack_conv(nn.Conv(ch, (fs, fs), (st, st), use_bias=False)(x))
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x[None, ...])[0]
            x = jax.nn.relu(x)

            x = self._stack_conv(nn.Conv(ch, (3, 3), use_bias=False)(x))
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x[None, ...])[0]
            x = jax.nn.relu(x)

            encoder_out.append(x)

            fs, st = 3, 2

        decoder_out = [x]

        for y in encoder_out[-2::-1]:

            ch = y.shape[-1]

            x = jax.image.resize(x, y.shape[:-1] + x.shape[-1:], "linear")
            x = jnp.concatenate([x, y], axis=-1)

            x = self._stack_conv(nn.Conv(ch, (3, 3), use_bias=False)(x))
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x[None, ...])[0]
            x = jax.nn.relu(x)

            x = self._stack_conv(nn.Conv(ch, (3, 3), use_bias=False)(x))
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x[None, ...])[0]
            x = jax.nn.relu(x)

            decoder_out.insert(0, x)

        return decoder_out



from __future__ import annotations

import jax
import jax.numpy as jnp
import optax

from xtrain import unpack_x_y_sample_weight

from ..ops import sub_pixel_samples, gather_patches, coords_of_patches
from .common import mean_over_boolean_mask, get_image_shape

def supervised_instance_loss(batch, prediction):
    """LACSS instance loss, supervised with segmentation label"""
    preds = prediction["predictions"]
    if not "segmentations" in preds:
        return 0

    inputs, labels, _ = unpack_x_y_sample_weight(batch)

    instance_mask = preds["segmentation_is_valid"]
    instance_logit = preds["segmentations"]
    n_patches = instance_logit.shape[0]

    if "gt_labels" in labels:  # labeled with image label
        gt_labels = labels["gt_labels"]
        if gt_labels.ndim == 2:
            gt_labels = gt_labels[None, :, :]
        locs = jnp.stack([ 
            preds["segmentation_z0_coord"],
            preds["segmentation_y0_coord"],
            preds["segmentation_x0_coord"],
        ], axis=-1)

        gt_patches = gather_patches(gt_labels, locs, instance_logit.shape[1:4])
        gt_patches = gt_patches == jnp.arange(1, n_patches+1)[:, None, None, None]
        gt_patches = gt_patches.astype("float32")

    elif "gt_masks" in labels:
        gt_segs = labels["gt_masks"][:n_patches]
        gt_bboxes = labels["gt_bboxes"][:n_patches]

        assert instance_logit.shape[1] == 1 or gt_segs.ndim == 4

        if gt_segs.ndim == 3:
            gt_segs = gt_segs[:, None, :, :]

        if gt_bboxes.shape[-1] == 4:
            gt_bboxes = jnp.c_[
                jnp.zeros([n_patches, 1], dtype=gt_bboxes.dtype), 
                gt_bboxes[:, :2], 
                jnp.ones([n_patches, 1], dtype=gt_bboxes.dtype),
                gt_bboxes[:, 2:],
            ]

        assert gt_segs.ndim == 4
        assert gt_bboxes.shape[-1] == 6
        assert gt_segs.shape[0] == n_patches, f"label length is {gt_segs.shape[0]}, but prediction length is {n_patches}"

        # compute rescaled coorinats in edge indexing
        (zc, yc, xc), cmask = coords_of_patches(preds, inputs['image'].shape[:-1])
        if "mask" in inputs:
            cmask = cmask & inputs["mask"][zc, yc, xc]
        z0, y0, x0, z1, y1, x1 = jnp.swapaxes(gt_bboxes, 0, 1)
        zc = zc + 0.5 - jnp.expand_dims(z0, (1,2,3))
        yc = yc + 0.5 - jnp.expand_dims(y0, (1,2,3))
        xc = xc + 0.5 - jnp.expand_dims(x0, (1,2,3))
        zc = zc / jnp.expand_dims(z1-z0, (1,2,3)) * gt_segs.shape[-3]
        yc = yc / jnp.expand_dims(y1-y0, (1,2,3)) * gt_segs.shape[-2]
        xc = xc / jnp.expand_dims(x1-x0, (1,2,3)) * gt_segs.shape[-1]

        # resample the label to match model coordinates
        gt_patches = jax.vmap(sub_pixel_samples)(
            gt_segs,
            jnp.stack([zc, yc, xc], axis=-1) -.5,
        ) >= 0.5
        gt_patches = jnp.where(cmask, gt_patches.astype("float32"), 0)

    else:
        return 0

    loss = optax.sigmoid_binary_cross_entropy(instance_logit, gt_patches)

    loss = loss.mean(axis=(1,2,3)).sum(where=instance_mask)

    return loss
    
    # return mean_over_boolean_mask(loss, instance_mask)

def instance_overlap_loss(batch, prediction, *, soft_label: bool = True):
    preds = prediction["predictions"]
    img_shape  = get_image_shape(batch)
    (cz, cy, cx), cmask = coords_of_patches(preds, img_shape)
    instances = jax.nn.sigmoid(preds['segmentations'])
    instances_sum = jax.numpy.zeros(img_shape)
    instance_mask = preds["segmentation_is_valid"]

    if soft_label:
        instances_sum = instances_sum.at[cz, cy, cx].add(jnp.where(cmask, instances, 0))
        loss = instances * jnp.where(cmask, instances_sum[cz, cy, cx] - instances, 1)
    else:
        log_yi = -jax.nn.log_sigmoid(-preds['segmentations'])
        instance_sum = instance_sum.at[cz, cy, cx].add(jnp.where(cmask, log_yi, 0))
        log_yi = jnp.where(
            cmask, 
            instance_sum[cz, cy, cx] - log_yi,
            1,
        )
        loss = instances * log_yi

    return mean_over_boolean_mask(loss, instance_mask)

def self_supervised_instance_loss(batch, prediction):
    """Unsupervised instance loss"""
    preds = prediction["predictions"]
    img_shape  = get_image_shape(batch)
    (cz, cy, cx), cmask = coords_of_patches(preds, img_shape)
    instances = jax.nn.sigmoid(preds['segmentations'])
    instance_mask = preds["segmentation_is_valid"]

    binary_mask = jax.lax.stop_gradient(jax.nn.sigmoid(preds["fg_pred"]))
    inputs, _, _ = unpack_x_y_sample_weight(batch)
    if 'mask' in inputs:
        binary_mask = binary_mask * inputs['mask']
    if binary_mask.ndim == 2:
        assert img_shape[0] == 1
        binary_mask = binary_mask[None, :, :]

    seg_patch = jnp.where(cmask, binary_mask[cz, cy, cx], 0)

    assert seg_patch.shape == instances.shape

    loss = (1 - seg_patch) * instances + seg_patch * (1 - instances)

    loss = mean_over_boolean_mask(loss, instance_mask)
    
    # loss += instance_overlap_loss(batch, prediction, soft_label=soft_label)

    return loss

def weakly_supervised_instance_loss(batch, prediction):
    """Instance loss supervised by image mask instead of instance masks"""
    preds = prediction["predictions"]
    img_shape  = get_image_shape(batch)
    (cz, cy, cx), cmask = coords_of_patches(preds, img_shape)
    instances = jax.nn.sigmoid(preds['segmentations'])
    instance_mask = preds["segmentation_is_valid"]

    loss = instance_overlap_loss(batch, prediction, soft_label=False)

    _, labels, _ = unpack_x_y_sample_weight(batch)
    if isinstance(labels, dict):
        seg = labels["gt_image_mask"].astype("float32")
    else:
        seg = labels.astype("float32")
    if seg.ndim == 2:
        assert img_shape[0] == 1
        seg = seg[None,:,:]
    seg_patch = jnp.where(cmask, seg[cz, cy, cx], 0)
    loss_ = (1.0 - seg_patch) * instances + seg_patch * (1.0 - instances)

    loss += mean_over_boolean_mask(loss_, instance_mask)    

    return loss

def segmentation_loss(batch, prediction, *, pretraining=False):
    _, labels, _ = unpack_x_y_sample_weight(batch)

    if labels is None:
        labels = {}

    if "gt_labels" in labels or "gt_masks" in labels:  # supervised
        return supervised_instance_loss(batch, prediction)
    elif "gt_image_mask" in labels:  # supervised by point + imagemask
        return weakly_supervised_instance_loss(batch, prediction)
    else:  # point-supervised
        return (
            self_supervised_instance_loss(batch, prediction)
            + instance_overlap_loss(batch, prediction, soft_label=not pretraining),
        )



from __future__ import annotations

import math
from typing import Sequence, Callable, Any

import flax.linen as nn
import jax
import jax.numpy as jnp
from flax.linen import normalization

from ..typing import ArrayLike, Array
from .common import DefaultUnpicklerMixin, ChannelAttention

class Segmentor3D(nn.Module, DefaultUnpicklerMixin):
    """LACSS 3D segmentation head.

    Attributes:
        n_layers: num of conv layers for feature mixing
        feature_scale: the spatail scale of the feature level
        instance_crop_size: Crop size for segmentation.
        pos_emb_shape: Dim of the learned position encoder.
    """
    n_layers: int = 1

    feature_scale: int = 4
    instance_crop_size: int = 96
    patch_dim: int = 48
    sig_dim: int = 1024
    pos_emb_shape: Sequence[int] = (8, 8, 8, 4)
    dtype: Any = None

    @property
    def patch_size(self):
        return self.instance_crop_size // self.feature_scale

    def _get_patch(self, feature, locations):
        patch_shape = (self.patch_size,) * 3
        locations = locations.astype(int) - jnp.asarray(patch_shape) // 2  # so we can use loc for indexing
        coords = jnp.mgrid[:self.patch_size, :self.patch_size, :self.patch_size] + locations[..., None, None, None] # N, 3, PS, PS, PS

        limit = jnp.expand_dims(jnp.asarray(feature.shape[:-1]), (1,2,3))

        valid_locs = (coords >= 0).all(axis=1) & (coords < limit).all(axis=1) # N, PS, PS
        patches = jnp.where(
            valid_locs[..., None],
            feature[coords[:, 0], coords[:, 1], coords[:, 2]],
            0,
        )
        return patches, locations

    def _get_signiture(self, patches):
        bw = self.patch_size // 4
        n, _, _, _, n_ch = patches.shape
        x = jax.image.resize(
            patches[:, bw:-bw, bw:-bw, bw:-bw, :],
            (n, 4, 4, 4, n_ch),
            "linear",
        ).reshape(n, -1)

        x = nn.Dense(self.sig_dim, dtype=self.dtype)(x)
        x = nn.LayerNorm(dtype=self.dtype)(x)

        return x


    def _add_pos_encoding(self, patches, sig_vec):
        x = sig_vec
        x = nn.gelu(nn.Dense(self.sig_dim, dtype=self.dtype)(x))
        x = nn.gelu(nn.Dense(self.sig_dim, dtype=self.dtype)(x))

        x = nn.Dense(math.prod(self.pos_emb_shape), dtype=self.dtype)(x)
        x = jax.image.resize(
            x.reshape(-1, *self.pos_emb_shape),
            (patches.shape[0], self.patch_size, self.patch_size, self.patch_size, self.pos_emb_shape[-1]),
            "linear",
        )

        encoding = nn.Dense(patches.shape[-1], use_bias=False, dtype=self.dtype)(x) #[n, ps, ps, ps, dim]

        # xa_dim = self.pos_emb_res * self.pos_emb_res * self.pos_emb_dim
        # xa = nn.Dense(xa_dim, dtype=self.dtype)(x)
        # xa = jax.image.resize(
        #     xa.reshape(-1, self.pos_emb_res, self.pos_emb_res, self.pos_emb_dim),
        #     (xa.shape[0], self.patch_size, self.patch_size, self.pos_emb_dim),
        #     "linear",
        # )
        # xa = nn.Dense(patches.shape[-1], use_bias=False, dtype=self.dtype)(xa) #[n, ps, ps, dim]

        # xb_dim = self.pos_emb_res * self.pos_emb_dim
        # xb = nn.Dense(xb_dim, dtype=self.dtype)(x)
        # xb = jax.image.resize(
        #     xb.reshape(-1, self.pos_emb_res, self.pos_emb_dim),
        #     (xa.shape[0], self.patch_size, self.pos_emb_dim),
        #     "linear",
        # )
        # xb = nn.Dense(patches.shape[-1], use_bias=False, dtype=self.dtype)(xb) #[n, ps, dim]

        # encoding = xa[:, None, :, :, :] + xb[:, :, None, None, :]

        patches = jax.nn.relu(patches + encoding)

        return patches


    @nn.compact
    def __call__(self, feature: ArrayLike, locations: ArrayLike) -> dict:
        """
        Args:
            features: [D, H, W, C] image feature
            locations: [N, 3]

        Returns:
            A nested dictionary of values representing segmentation outputs.
              * segmentor/logits: all segment logits
              * predictions/segmentations: logits of designated class [N, D, ps, ps]
              * predictions/segmentation_y0_coord: y coord of patch top-left corner [N]
              * predictions/segmentation_x0_coord: x coord of patch top-left corner [N]
              * predictions/segmentation_is_valid: mask for valid patches [N]
        """
        locations = locations / self.feature_scale

        x = feature
        dim = self.patch_dim
        for _ in range(self.n_layers):
            x = nn.Conv(dim, (3,3,3), dtype=self.dtype)(x)
            x = nn.gelu(x)
        x = nn.Conv(dim, (3,3,3), dtype=self.dtype)(x)

        patches, patch_locs = self._get_patch(x, locations) # N, PS, PS, PS, ch

        patches = ChannelAttention(dtype=self.dtype)(patches)

        patch_sigs = self._get_signiture(patches)

        patches = self._add_pos_encoding(patches, patch_sigs)

        logits = nn.ConvTranspose(1, (2, 2, 2), strides=(2, 2, 2), dtype=self.dtype)(patches)

        output_shape = (locations.shape[0],) + (self.instance_crop_size,) * 3
        logits = jax.image.resize(logits.squeeze(-1), output_shape, "linear")
        instance_mask = locations[:, -1] >= 0
        patch_locs = jnp.where(
            instance_mask[:, None],
            patch_locs * self.feature_scale,
            -1,
        )

        return dict(
            segmentor=dict(
                patch_signitures=patch_sigs,
            ),
            predictions=dict(
                segmentations=logits,
                segmentation_z0_coord=patch_locs[:, 0],
                segmentation_y0_coord=patch_locs[:, 1],
                segmentation_x0_coord=patch_locs[:, 2],
                segmentation_is_valid=instance_mask,
            ),
        )



import importlib.machinery
import logging
import traceback
import threading
import types
from concurrent import futures
from pathlib import Path

import grpc
import numpy as np
import torch
import typer

from . import proto
from .common import decode_image, TokenValidationInterceptor

app = typer.Typer(pretty_exceptions_enable=False)

_MAX_MSG_SIZE=1024*1024*128


def process_input(request: proto.DetectionRequest):
    pixels = request.image_data.pixels
    settings = request.detection_settings

    image = decode_image(pixels)

    return image


def process_result(masks, image):
    import cv2
    from skimage.measure import regionprops

    response = proto.DetectionResponse()

    for rp in regionprops(masks):
        mask = rp.image.astype("uint8")
        c, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        c = np.array(c[0], dtype=float).squeeze(1)
        c = c + np.array([rp.bbox[1] , rp.bbox[0]])
        c = c - 0.5

        scored_roi = proto.ScoredROI(
            score = 1.0,
            roi = proto.ROI(
                polygon = proto.Polygon(points = [proto.Point(x=p[0], y=p[1]) for p in c]),
            )
        )

        response.detections.append(scored_roi)

    return response


class SegformerServicer(proto.LacssServicer):

    def __init__(self, mod, gpu, model_path, model_path2):
        self.mod = mod
        self.model_path = model_path
        self.model_path2 = model_path2
        self.device = "cuda:0" if gpu else "cpu"
        self._lock = threading.RLock()

    def predict(self, img_data):
        hflip_tta = self.mod.HorizontalFlip()
        vflip_tta = self.mod.VerticalFlip()

        if img_data.shape[-1] == 1:
            img_data = np.repeat(img_data, 3, axis=-1)
        elif img_data.shape[-1] == 2:
            raise ValueError("Model accept either 1-channel or 3-channel images.")

        img_data = self.mod._normalize(img_data)
        img_data = np.moveaxis(img_data, -1, 0)

        model = torch.load(self.model_path, map_location=self.device)
        model.eval()

        img_data = img_data.to(self.device)
        img_size = img_data.shape[-1] * img_data.shape[-2]
        
        if img_size < 1150000 and 900000 < img_size:
            overlap = 0.5
        else:
            overlap = 0.6

        with torch.no_grad():
            img0 = img_data
            outputs0 = self.mod.sliding_window_inference(
                img0,
                512,
                4,
                model,
                padding_mode="reflect",
                mode="gaussian",
                overlap=overlap,
                device="cpu",
            )
            outputs0 = outputs0.cpu().squeeze()

            if img_size < 2000 * 2000 or img_size > 5000 * 5000:
                
                model.load_state_dict(torch.load(self.model_path2, map_location=self.device))
                model.eval()
                
                img2 = hflip_tta.apply_aug_image(img_data, apply=True)
                outputs2 = self.mod.sliding_window_inference(
                    img2,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs2 = hflip_tta.apply_deaug_mask(outputs2, apply=True)
                outputs2 = outputs2.cpu().squeeze()

                outputs = torch.zeros_like(outputs0)
                outputs[0] = (outputs0[0] + outputs2[0]) / 2
                outputs[1] = (outputs0[1] - outputs2[1]) / 2
                outputs[2] = (outputs0[2] + outputs2[2]) / 2
                
            else:
                # Hflip TTA
                img2 = hflip_tta.apply_aug_image(img_data, apply=True)
                outputs2 = self.mod.sliding_window_inference(
                    img2,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs2 = hflip_tta.apply_deaug_mask(outputs2, apply=True)
                outputs2 = outputs2.cpu().squeeze()
                img2 = img2.cpu()
                
                ##################
                #                #
                #    ensemble    #
                #                #
                ##################
                
                model.load_state_dict(torch.load(args.model_path2, map_location=args.device))
                model.eval()
                
                img1 = img_data
                outputs1 = self.mod.sliding_window_inference(
                    img1,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs1 = outputs1.cpu().squeeze()
                
                # Vflip TTA
                img3 = vflip_tta.apply_aug_image(img_data, apply=True)
                outputs3 = self.mod.sliding_window_inference(
                    img3,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs3 = vflip_tta.apply_deaug_mask(outputs3, apply=True)
                outputs3 = outputs3.cpu().squeeze()
                img3 = img3.cpu()

                # Merge Results
                outputs = torch.zeros_like(outputs0)
                outputs[0] = (outputs0[0] + outputs1[0] + outputs2[0] - outputs3[0]) / 4
                outputs[1] = (outputs0[1] + outputs1[1] - outputs2[1] + outputs3[1]) / 4
                outputs[2] = (outputs0[2] + outputs1[2] + outputs2[2] + outputs3[2]) / 4
                
            pred_mask = self.mod.post_process(outputs.squeeze(0).cpu().numpy(), self.device)

            return pred_mask


    def RunDetection(self, request, context):
        with self._lock:

            logging.info(f"Received message of size {request.ByteSize()}")

            try:
                image = process_input(request)

                if image.ndim == 4:
                    raise ValueError("Model does not support 3D input")

                logging.info(f"received image {image.shape}")

                preds = self.predict(image)

                response = process_result(preds, image)

                logging.info(f"Reply with message of size {response.ByteSize()}")

                return response
            
            except ValueError as e:
                
                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(grpc.StatusCode.INVALID_ARGUMENT, repr(e))

            except Exception as e:

                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}")



    def RunDetectionStream(self, request_iterator, context):
        with self._lock:
            request = proto.DetectionRequest()

            for next_request in request_iterator:

                if next_request.image_data.HasField("pixels"):
                    request.image_data.pixels.CopyFrom(next_request.image_data.pixels)

                if next_request.image_data.HasField("image_annotation"):
                    request.image_data.image_annotation.CopyFrom(next_request.image_data.image_annotation)
                
                if next_request.HasField("detection_settings"):
                    request.detection_settings.CopyFrom(next_request.detection_settings)
                
                if request.image_data.HasField("pixels"):
                    yield self.RunDetection(request, context)


@app.command()
def main(
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool|None = None,
    debug: bool = False,
    compression: bool = True,
    gpu: bool = True,
    predict_py_file: Path = "./predict.py",
    model_path : Path = "./main_model.pt",
    model_path2 : Path = "./sub_model.pt"
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    print ("server starting ...")

    loader = importlib.machinery.SourceFileLoader('predict', predict_py_file)
    mod = types.ModuleType( loader.name )
    loader.exec_module(mod)

    if token is None:
        token = not local
    if token:
        import secrets

        token_str = secrets.token_urlsafe(64)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token_str}")
        print("=======================================================================")
        print()
    else:
        token_str = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip if compression else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token_str),),
        options=(("grpc.max_receive_message_length", _MAX_MSG_SIZE),),
    )

    proto.add_LacssServicer_to_server(
        SegformerServicer(mod, gpu, model_path, model_path2), 
        server,
    )

    if local:
        server.add_secure_port(f"127.0.0.1:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logging.info(f"lacss_server: listening on port {port}")

    print ("server starting ... ready")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    app()



""" 
Attributes:
    model_urls: URLs for build-in pretrain models. e.g model_urls["default"].
"""
from __future__ import annotations

import logging
import math
import time
from functools import reduce, partial
from typing import Mapping, Sequence, Tuple

import cv2
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np

from ..ops import patches_to_label, non_max_suppression, bboxes_of_patches, crop_and_resize_patches, box_iou_similarity, sub_pixel_samples
from ..utils import load_from_pretrained
from ..typing import Array, ArrayLike
from ..data.utils import image_standardization

Shape = Sequence[int]

def _remove_edge_instances(
    pred, pos, img_sz,
    gs: int = 1024+64,
):
    bboxes = pred['pred_bboxes']
    removal = np.zeros([bboxes.shape[0]], dtype=bool)

    dim = len(img_sz)

    removal |= ((bboxes[:, :dim] <= 0) & (pos > 0)).any(axis=-1)
    removal |= ((bboxes[:, dim:] >= gs) & (pos + gs < img_sz)).any(axis=-1)

    pred = {k:v[~removal] for k, v in pred.items()}

    return pred

def _get_contour(mask, chain_approx=cv2.CHAIN_APPROX_SIMPLE):
    """ returns an edge index polygon representing the outer contour """
    c, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, chain_approx)
    max_len_element = reduce(
        lambda a, b: a if len(a) >= len(b) else b,
        c,
        np.zeros([0, 1, 2], dtype=int),
    )
    return max_len_element.squeeze(1).astype(float) + .5
    
def _to_polygons(
    predictions, mask,
    *,
    segmentation_threshold=0.0,
    chain_approx=cv2.CHAIN_APPROX_SIMPLE,
) -> list:
    polygons = []

    y0s = np.asarray(predictions["segmentation_y0_coord"])
    x0s = np.asarray(predictions["segmentation_x0_coord"])
    segs = np.asarray(predictions["segmentations"] >= segmentation_threshold).astype("uint8")
    mask = np.array(mask) 

    assert segs.shape[1] == 1
    segs = segs.squeeze(1)

    for k in range(segs.shape[0]):
        if mask[k]:
            polygon = _get_contour(segs[k], chain_approx) + [x0s[k], y0s[k]]
            if len(polygon) > 0:
                polygons.append(polygon)
            else:
                mask[k] = False

    return polygons, mask


def _get_mesh(seg):
    from skimage.measure import marching_cubes
    try:
        verts, faces, norms, _ = marching_cubes(seg, allow_degenerate=False)
        return dict(verts=verts, faces=faces, norms=norms)
    except:
        return None

def _to_mesh(
    predictions, mask,
    *,
    segmentation_threshold=0,
):
    from skimage.measure import marching_cubes
    meshes = []

    z0s = np.asarray(predictions["segmentation_z0_coord"])
    y0s = np.asarray(predictions["segmentation_y0_coord"])
    x0s = np.asarray(predictions["segmentation_x0_coord"])
    segs = np.asarray(predictions["segmentations"] >= segmentation_threshold).astype("uint8")
    mask = np.array(mask) 

    for k in range(segs.shape[0]):
        if mask[k]:
            mesh = _get_mesh(segs[k])
            if mesh:
                mesh['verts'] += [z0s[k], y0s[k], x0s[k]]
                meshes.append(mesh)
            else:
                mask[k] = False

    return meshes, mask  

def _format_image(image, target_shape, normalize):
    from skimage.transform import resize

    if image.shape[:-1] != tuple(target_shape):
        image = resize(image, target_shape)

    is2d = image.ndim == 3

    if normalize:
        image = image - image.mean()
        image = image / (image.std() + 1e-8)

    if image.shape[-1] == 1:
        image = np.repeat(image, 3, axis=-1)

    assert image.shape[-1] <= 3, f"input image has more than 3 channels"

    padding_shape = (target_shape - 1) // 32 * 32 + 32

    if is2d:
        if (padding_shape <=  544).all():
            padding_shape[:] = 544
        elif (padding_shape <=  1088).all():
            padding_shape[:] = 1088
    else:
        padding_shape[:] = padding_shape.max()
        if (padding_shape <  256).all():
            padding_shape[:] = 256
        else:
            padding_shape[:] = padding_shape.max()
    
    padding = [ [0, s-s0] for s, s0 in zip(padding_shape, image.shape[:-1])]
    padding += [[0, 3-image.shape[-1]]]
    image = np.pad(image, padding)

    return image


def _nms(preds, th):
    from lacss.ops import box_iou_similarity
    scores = preds["pred_scores"]
    asort = np.argsort(scores)[::-1]
    preds = {k: v[asort] for k, v in preds.items()}

    boxes = preds['pred_bboxes']
    sm = box_iou_similarity(boxes, boxes)
    for k in range(sm.shape[0]):
        sm[k, :k+1] = -1

    supressed = np.zeros([sm.shape[0]], dtype=bool)
    for k in range(sm.shape[0]):
        if not supressed[k]:
            supressed |= sm[k] > th
    
    preds = {k: v[~supressed] for k, v in preds.items()}

    return preds


def _predict_partial(model, params, image):
    model = model.copy()
    model.segmentor = None
    model.segmentor_3d = None

    lpn_out, seg_features = [], []
    for p in params:
        model_out =  model.apply(dict(params=p), image)
        lpn_out.append( model_out['detector'] )
        seg_features.append(model_out['seg_features'])

    lpn_out = jax.tree_util.tree_map(
        lambda *x: jnp.stack(x), * lpn_out
    )
    lpn_out = dict(
        logits = lpn_out['logits'].mean(axis=0),
        regressions =  lpn_out['regressions'].mean(axis=0),
        ref_locs = lpn_out['ref_locs'][0],
    )
    lpn_out['pred_locs'] = lpn_out['ref_locs'] + lpn_out['regressions'] * model.detector.feature_scale

    return seg_features, lpn_out


@partial(jax.jit, static_argnums=0)
def _predict(model, params, image):
    from lacss.modules.lpn import generate_predictions

    seg_features, lpn_out = _predict_partial(model, params, image)

    predictions = generate_predictions(model.detector, lpn_out)

    assert len(seg_features) == len(params)

    seg_predictions = []
    if image.ndim == 3 and model.segmentor is not None:
        for x, p in zip(seg_features, params):
            seg_predictions.append( 
                model.segmentor.apply(
                    dict(params=p['segmentor']), 
                    x, 
                    predictions['locations']
                )['predictions'] 
            )

    elif image.ndim == 4 and model.segmentor_3d is not None:
        seg_predictions = []
        for x, p in zip(seg_features, params):
            seg_predictions.append( 
                model.segmentor_3d.apply(
                    dict(params=p['segmentor_3d']), 
                    x, 
                    predictions['locations']
                )['predictions'] 
            )

    if len(seg_predictions) > 0:
        predictions['segmentations'] = jnp.mean(
            jnp.array([ x['segmentations'] for x in seg_predictions]),
            axis = 0,
        )

        seg_prediction = seg_predictions[0]
        del seg_prediction['segmentations']

        predictions.update( seg_prediction )
        
    return predictions


class Predictor:
    """Main class interface for model deployment. This is the only class you
    need if you don't train your own model

    Examples:
        The most common use case is to use a build-in pretrained model.

            import lacss.deploy

            # look up the url of a build-in mode
            url = lacss.deploy.model_urls["default"]

            # create the predictor instance
            predictor = lacss.deploy.Predictor(url)

            # make a prediction
            label = predictor.predict(image)

    Attributes:
        module: The underlying FLAX module
        params: Model weights.
    """
    def __init__(
            self, 
            url: str | tuple[nn.Module, dict],
            *,
            grid_size = 544,
            step_size = 480,
        ):
        """Construct Predictor

        Args:
            url: A URL or local path to the saved model.
                URLs for build-in pretrained models can be found in lacss.deploy.model_urls
        """

        if isinstance(url, tuple) and len(url) == 2:
            if not isinstance(url[0], nn.Module):
                raise ValueError(
                    "Initiaize the Predictor with a tuple, but the first element is not a Module."
                )

            module, params = url

        else:
            module, params = load_from_pretrained(url)


        if isinstance(params, Mapping):
            params = [params]

        self.module = module
        self.params = params

        assert step_size < grid_size, f"step_size ({step_size}) not smaller than grid_size ({grid_size})"
        assert grid_size % 32 == 0, f"grid_size ({grid_size}) is not divisable by 32"

        self.gs = grid_size
        self.ss = step_size


    def predict(
        self,
        image: ArrayLike,
        *,
        output_type: str = "label",
        reshape_to: int|tuple[int]|None = None,
        min_area: float = 0,
        score_threshold: float = 0.5,
        segmentation_threshold: float = 0.5,
        nms_iou: float = 1,
        normalize: bool = True,
        remove_out_of_bound: bool|None = None,
    ) -> dict:
        """Predict segmentation.

        Args:
            image: A ndarray of (h,w,c) or (d,h,w,c) format. c must be 1-3

        Keyword Args:
            output_type: "label" | "contour" | "bbox"
            reshape_to: If not None, the input image will be resized internally before send
                to the model. The results will be resized back to the scale of the orginal input image.
            min_area: Minimum area of a valid prediction.
            score_threshold: Min score needed to be included in the output.
            segmentation_threshold: Threshold value for segmentation
            nms_iou: IOU threshold value for non-max-supression post-processing

        Returns:
            For "label" output:

                - pred_scores: The prediction scores of each instance.
                - pred_label: a 2D image label. 0 is background.

            For "contour" output:

                - pred_scores: The prediction scores of each instance.
                - pred_contours: a list of polygon arrays in x-y format.

            For "bbox" output (ie MaskRCNN):

                - pred_scores: The prediction scores of each instance.
                - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
                - pred_masks:  A 3d array representing (rescaled) segmentation mask within bboxes
        """
        image = np.asarray(image)
        
        logging.debug(f"started prediction with image {image.shape}")
        start_time = time.time()

        if remove_out_of_bound is not None:
            import warnings
            warnings.warn("remove_out_of_bound is deprecated", DeprecationWarning, 2)

        if image.ndim == 2 or image.shape[-1] > 3:
            logging.warning("input seems to have no channel dim. Add one")
            image = image[..., None]

        orig_shape = image.shape[:-1]
        assert image.ndim == 3 or image.ndim == 4, f"illegal image dim {image.shape}"
        is_3d = image.ndim == 4

        if not output_type in ("bbox", "label", "contour", "_raw"):
            raise ValueError(
                f"output_type should be 'bbox'|'label'|'contour'. Got {output_type} instead."
            )

        # if output_type == "contour" and image.ndim == 4:
        #     raise ValueError(f"contour output is only supported on 2D inputs")
        
        if reshape_to is None: 
            reshape_to = np.array(orig_shape)
        else:
            reshape_to = np.broadcast_to(reshape_to, [len(orig_shape)])
        scaling = reshape_to / orig_shape

        image = _format_image(image, reshape_to, normalize=normalize)

        seg_logit_threshold = math.log(segmentation_threshold / (1 - segmentation_threshold))

        logging.debug(f"done preprocessing")

        preds = _predict(self.module, self.params, image)

        scores = preds["scores"]
        bboxes = bboxes_of_patches(
            preds, 
            threshold=seg_logit_threshold, 
            image_shape=reshape_to,
        )

        elapsed = (time.time() - start_time) * 1000
        logging.debug(f"finished model execution in {elapsed:.2f} ms")

        # mark all invalid instances
        instance_mask = np.asarray(preds["segmentation_is_valid"])
        instance_mask &= scores >= score_threshold
        # instance_mask &= (bboxes >= 0).all(axis=-1)
        if min_area > 0:
            areas = jnp.count_nonzero(
                preds["segmentations"] > segmentation_threshold, axis=(1, 2, 3)
            )
            instance_mask &= areas >= min_area * np.prod(scaling)
        if nms_iou > 0 and nms_iou < 1:
            sel = non_max_suppression(
                scores, bboxes, bboxes.shape[0],
                threshold=nms_iou,
                return_selection=True,
                similarity_func=box_iou_similarity,
            )
            instance_mask = instance_mask & sel

        if output_type == "_raw":
            return dict(
                pred_scores=np.asarray(scores)[instance_mask],
                pred_locations=np.asarray(preds["locations"])[instance_mask] / scaling,
                pred_bboxes=(np.asarray(bboxes) / np.r_[scaling, scaling])[instance_mask],
                pred_segmentations=np.asarray(preds["segmentations"])[instance_mask],
                pred_segmentation_pos=np.stack([
                    preds["segmentation_z0_coord"],
                    preds["segmentation_y0_coord"],
                    preds["segmentation_x0_coord"],
                ], axis=-1)[instance_mask],
            )
        elif output_type == "bbox":
            target_shape = np.broadcast_to([36], [len(orig_shape)])
            segmentations = crop_and_resize_patches(
                preds, bboxes,
                target_shape=target_shape,
                convert_logits=True,
            )
            results = dict(
                pred_scores=np.asarray(scores)[instance_mask],
                pred_bboxes=(np.asarray(bboxes) / np.r_[scaling, scaling])[instance_mask],
                pred_masks=np.asarray(segmentations)[instance_mask],
                pred_locations=np.asarray(preds["locations"])[instance_mask] / scaling,
            )

        elif output_type == "contour":
            if not is_3d:
                polygons, instance_mask = _to_polygons(
                    preds, instance_mask, segmentation_threshold=seg_logit_threshold,
                )
                polygons = [(c + 0.5) / scaling[::-1] for c in polygons]

                results = dict(
                    pred_scores=np.asarray(scores)[instance_mask],
                    pred_bboxes=(np.asarray(bboxes) / np.r_[scaling, scaling])[instance_mask],
                    pred_contours=polygons,
                )
            else:
                meshes, instance_mask = _to_mesh(
                    preds, instance_mask,
                    segmentation_threshold=seg_logit_threshold,
                )

                for mesh in meshes:
                    mesh['verts'] /=  scaling

                results = dict(
                    pred_scores=np.asarray(scores)[instance_mask],
                    pred_bboxes=(np.asarray(bboxes) / np.r_[scaling, scaling])[instance_mask],
                    pred_contours=meshes,
                )                

        else:  # Label
            label = patches_to_label(
                preds, reshape_to,
                mask=instance_mask,
                score_threshold=0,
                threshold=seg_logit_threshold,
            )
            if not (scaling == 1.0).all():
                label = jax.image.resize(
                    label, orig_shape, "nearest",
                )

            results = dict(
                pred_scores=np.asarray(scores)[instance_mask],
                pred_label=np.asarray(label),
            )

        elapsed = (time.time() - start_time) * 1000
        logging.debug(f"done formating output in {elapsed:.2f} ms")

        return results

    def predict_on_large_image(
        self,
        image: ArrayLike,
        *,
        reshape_to: tuple[int,...]|None = None,
        gs: int|None = None,
        ss: int|None = None,
        nms_iou: float = 0.0,
        score_threshold: float = 0.5,
        segmentation_threshold: float = 0.5,
        min_area: float = 0,
        min_cells_per_patch: int = 0,
        output_type: str = "label",
    ) -> dict:
        """Make prediction on very large image by dividing into a grid.

        Direct model prediction on large image may cause out-of-memory error. This
        method divided the large image to smaller grid and then stitch the results
        to form the complete prediction.

        Args:
            image: An image with (H, W, C) format.
            gs: An int value. Grid size of the computation.
            ss: An int value of stepping size. Must be small than gs to produce valid
                results.

        Keyword Args:
            min_area: Minimum area of a valid prediction.
            score_threshold: Min score needed to be included in the output.
            segmentation_threshold: Threshold value for segmentation.
            nms_iou: IOU threshold value for non-max-supression durint post-processing

        Returns:
            - pred_scores: The prediction scores of each instance.
            - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
            - pred_masks: A 3d array representing segmentation within bboxes
        """
        if not output_type in ("bbox", "label", "contour"):
            raise ValueError(
                f"output_type should be 'bbox'|'label'|'contour'. Got {output_type} instead."
            )

        if image.ndim == 2 or image.shape[-1] > 3:
            logging.warning("input seems to have no channel dim. Add one")
            image = image[..., None]

        orig_shape = image.shape[:-1]
        assert image.ndim == 3 or image.ndim == 4, f"ilegal image dim {image.shape}"

        if ss is None:
            ss = 1024 if image.ndim == 3 else 192
            gs = None
        if gs is None:
            gs = ss + 64 if image.ndim == 3 else 32

        if reshape_to is None: 
            reshape_to = np.array(orig_shape)
        else:
            reshape_to = np.broadcast_to(reshape_to, [len(orig_shape)])
        scaling = reshape_to / orig_shape
        
        image = jax.image.resize(image, tuple(reshape_to) + (image.shape[-1],), "linear")
        image = image - image.mean()
        image = image / (image.std() + 1e-8)

        grid_positions = [slice(0, max(d-(gs-ss), 1), ss) for d in reshape_to]
        grid_positions = np.moveaxis(np.mgrid[grid_positions], 0, -1)
        grid_positions = grid_positions.reshape(-1, len(orig_shape))

        preds = []
        for pos in grid_positions:
            slices = (slice(x, x + gs) for x in pos)
            patch = image.__getitem__(tuple(slices))

            logging.info(f"Processing grid {pos} with patch size {patch.shape}")

            pred = self.predict(
                patch,
                score_threshold=score_threshold,
                segmentation_threshold=segmentation_threshold,
                min_area=min_area,
                nms_iou=0.5,
                normalize=False,
                output_type="bbox",
            )
            pred = _remove_edge_instances(
                pred, pos, reshape_to,
                gs=gs,
            )

            if len(pred['pred_scores']) > min_cells_per_patch:
                preds.append(dict(
                    pred_scores = pred['pred_scores'],
                    pred_bboxes = pred['pred_bboxes'] + np.r_[pos, pos],
                    pred_masks = pred['pred_masks'],
                    # pred_locations = pred["pred_locations"] + pos
                ))

        preds = jax.tree_util.tree_map(lambda *x: np.concatenate(x), *preds)

        # nms
        logging.info(f"nms...")

        # asort = np.argsort(preds["pred_scores"])[::-1]
        # preds = jax.tree_util.tree_map(lambda x: x[asort], preds)
        preds = _nms(preds, nms_iou)

        # rescale
        preds["pred_bboxes"] /= np.r_[scaling, scaling]
        # preds["pred_locations"] /= scaling

        if output_type == "bbox":
            return preds

        if output_type == "contour":
            if len(orig_shape) == 2:
                polygons = []
                _, sy, sx = preds['pred_masks'].shape
                for box, seg in zip(preds['pred_bboxes'], preds['pred_masks']):
                    y0, x0, y1, x1 = box
                    polygon = _get_contour((seg >= 0.5).astype("uint8"))
                    polygon = polygon / [sx, sy] * [x1-x0, y1-y0] + [x0, y0]
                    polygons.append(polygon)
                preds['pred_contours'] = polygons

            else:
                meshes = []
                _, sz, sy, sx = preds['pred_masks'].shape
                for box, seg in zip(preds['pred_bboxes'], preds['pred_masks']):
                    z0, y0, x0, z1, y1, x1 = box
                    mesh = _get_mesh((seg >= 0.5).astype("uint8"))
                    assert mesh is not None
                    mesh['verts'] = mesh['verts'] / [sz, sy, sx] * [z1-z0, y1-y0, x1-x0] + [z0, y0, x0]
                    meshes.append(mesh)

                preds['pred_contours'] = meshes

            return preds

        else: # label
            from scipy.interpolate import interpn
            logging.info(f"Generating label...")
            label = np.zeros(orig_shape, dtype=int)
            bboxes = preds['pred_bboxes']
            masks = preds['pred_masks']
            dim = bboxes.shape[-1] // 2

            if dim == 2:
                _, sy, sx = masks.shape
                mask_grid = (np.arange(sy), np.arange(sx))
            else:
                _, sz, sy, sx = masks.shape
                mask_grid = (np.arange(sz), np.arange(sy), np.arange(sx))

            p0s = np.maximum(0, np.floor(bboxes[:, :dim]).astype(int))
            p1s = np.minimum(np.ceil(bboxes[:, dim:]).astype(int), label.shape)

            for label_idx, (p0, p1, mask) in enumerate(zip(p0s, p1s, masks)):
                slices = [slice(a,b) for a,b in zip(p0, p1)]
                coords = tuple(np.mgrid[slices])
                src_coords = (np.stack(coords, axis=-1) + 0.5 - p0) / (p1-p0) * mask.shape - 0.5
                values = interpn(mask_grid, mask, src_coords, bounds_error=False, fill_value=0)
                values = (values >= segmentation_threshold ) * (label_idx + 1)
                label[coords] = np.where(
                    label[coords] == 0,
                    values,
                    label[coords],
                )
            preds['pred_label'] = label

            return preds



from __future__ import annotations

from functools import partial
from typing import Callable, Sequence, Any

import flax.linen as nn

import jax

from .common import DefaultUnpicklerMixin, ChannelAttention, picklable_relu
from ..typing import Array, ArrayLike


class FPN(nn.Module, DefaultUnpicklerMixin):
    out_channels: int = 384
    se_ratio: int = 0
    activation: Callable[[Array], Array] = picklable_relu
    dtype: Any = None

    @nn.compact
    def __call__(self, inputs: Sequence[ArrayLike]) -> Sequence[Array]:
        out_channels = self.out_channels

        outputs = [self.activation(nn.Dense(out_channels, dtype=self.dtype)(x)) for x in inputs]

        if self.se_ratio > 0:
            outputs = [ChannelAttention(self.se_ratio, dtype=self.dtype)(x) for x in outputs]

        for k in range(len(outputs) - 1, 0, -1):
            x = jax.image.resize(outputs[k], outputs[k - 1].shape, "nearest")
            x += outputs[k - 1]
            x = nn.Conv(out_channels, (3, 3), dtype=self.dtype)(x)
            x = self.activation(x)
            outputs[k - 1] = x

        return outputs


class ConvIntegrator(nn.Module, DefaultUnpicklerMixin):
    """A UNet-like feature integrator.

    Unlike FPN, its feature spaces are not consistent at different spatial scales.
    high-res scale should have lower number of features (like unet)
    """

    n_features: Sequence[int] | None = None
    n_layers: int = 1
    norm: Callable = nn.GroupNorm
    activation: Callable = nn.gelu

    @nn.compact
    def __call__(self, x: Sequence[Array]) -> Sequence[Array]:
        n_features = self.n_features or [f.shape[-1] for f in x]

        y = x[-1]
        for _ in range(self.n_layers):
            y = nn.Conv(n_features[-1], (3, 3))(y)
            y = self.activation(y)
        output = [y]

        for k in range(-1, -1 * len(x), -1):
            y = nn.ConvTranspose(
                x[k - 1].shape[-1],
                (3, 3),
                strides=(2, 2),
            )(output[0])

            y = self.norm()(x[k - 1] + y)

            for _ in range(self.n_layers):
                y = nn.Conv(n_features[k - 1], (3, 3))(y)
                y = self.activation(y)

            output.insert(0, y)

        return output



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: biopb/lacss/image_data.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'biopb/lacss/image_data.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from biopb.lacss import roi_pb2 as biopb_dot_lacss_dot_roi__pb2
from biopb.lacss import bindata_pb2 as biopb_dot_lacss_dot_bindata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1c\x62iopb/lacss/image_data.proto\x12\x0b\x62iopb.lacss\x1a\x15\x62iopb/lacss/roi.proto\x1a\x19\x62iopb/lacss/bindata.proto\"\xa6\x03\n\x06Pixels\x12%\n\x07\x62indata\x18\x03 \x01(\x0b\x32\x14.biopb.lacss.BinData\x12\x17\n\x0f\x64imension_order\x18\x04 \x01(\t\x12\r\n\x05\x64type\x18\x05 \x01(\t\x12\x0e\n\x06size_x\x18\t \x01(\r\x12\x0e\n\x06size_y\x18\n \x01(\r\x12\x0e\n\x06size_z\x18\x0b \x01(\r\x12\x0e\n\x06size_c\x18\x0c \x01(\r\x12\x0e\n\x06size_t\x18\r \x01(\r\x12\x1c\n\x0fphysical_size_x\x18\x0e \x01(\x02H\x00\x88\x01\x01\x12\x1c\n\x0fphysical_size_y\x18\x0f \x01(\x02H\x01\x88\x01\x01\x12\x1c\n\x0fphysical_size_z\x18\x10 \x01(\x02H\x02\x88\x01\x01\x12\x1a\n\x12physical_size_unit\x18\x11 \x01(\t\x12\x1b\n\x0etime_increment\x18\x12 \x01(\x02H\x03\x88\x01\x01\x12\x1b\n\x13time_increment_unit\x18\x15 \x01(\tB\x12\n\x10_physical_size_xB\x12\n\x10_physical_size_yB\x12\n\x10_physical_size_zB\x11\n\x0f_time_increment\"1\n\x0fImageAnnotation\x12\x1e\n\x04rois\x18\x01 \x03(\x0b\x32\x10.biopb.lacss.ROI\"h\n\tImageData\x12#\n\x06pixels\x18\x01 \x01(\x0b\x32\x13.biopb.lacss.Pixels\x12\x36\n\x10image_annotation\x18\x02 \x01(\x0b\x32\x1c.biopb.lacss.ImageAnnotationB\x0f\n\x0b\x62iopb.lacssP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'biopb.lacss.image_data_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\013biopb.lacssP\001'
  _globals['_PIXELS']._serialized_start=96
  _globals['_PIXELS']._serialized_end=518
  _globals['_IMAGEANNOTATION']._serialized_start=520
  _globals['_IMAGEANNOTATION']._serialized_end=569
  _globals['_IMAGEDATA']._serialized_start=571
  _globals['_IMAGEDATA']._serialized_end=675
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from functools import partial
from typing import Sequence, Tuple, Callable

import flax.linen as nn
import jax
import jax.numpy as jnp

from .lpn import generate_predictions
from .common import DefaultUnpicklerMixin
from ..typing import Array, ArrayLike, Any

class LPN3D(nn.Module, DefaultUnpicklerMixin):
    """Location 3d detection head

    Attributes:
        n_layers: num of conv layers for feature mixing
        nms_threshold: non-max-supression threshold, if performing nms on detected locations.
        pre_nms_topk: max number of detections to be processed regardless of nms, ignored if negative
        max_output: number of detection outputs 
    """

    # network hyperparams
    n_layers: int = 2
    dim: int = 384
    feature_scale: int = 4
    dtype: Any = None

    # detection hyperparams
    nms_threshold: float = 8.0
    pre_nms_topk: int = -1
    max_output: int = 256
    min_score: float = 0.2

    def _block(self, feature: Array) -> dict:
        x = feature

        depth, height, width = x.shape[:-1]

        logits = nn.Conv(1, (1, 1), dtype=self.dtype)(x)
        regressions = nn.Conv(3, (1, 1), dtype=self.dtype)(x)

        ref_locs = jnp.moveaxis(jnp.mgrid[:depth, :height, :width] + 0.5, 0, -1) * self.feature_scale
        locs = ref_locs + regressions * self.feature_scale

        return dict(
            regressions=regressions.reshape(-1, 3),
            logits = logits.reshape(-1),
            ref_locs=ref_locs.reshape(-1, 3),
            pred_locs=locs.reshape(-1, 3),
        )


    def _mix_feaures(self, feature: Array) -> Array:
        x = feature
        dim = self.dim
        for _ in range(self.n_layers):
            x = nn.Conv(dim, (3,3,3), dtype=self.dtype)(x)
            x = nn.gelu(x)

        return x

    @nn.compact
    def __call__(self, feature: ArrayLike, mask: ArrayLike|None = None) -> dict:
        feature = jnp.asarray(feature)

        x = self._mix_feaures(feature)
        x = self._block(x)
        predictions = generate_predictions(self, x)

        if mask is not None:
            pred_locs = jnp.floor(predictions["locations"]).astype(int)
            pred_locs_ = jnp.clip(pred_locs, 0, jnp.array(mask.shape) - 1)
            masked = jnp.where(
                (pred_locs == pred_locs_).all(axis=1), 
                mask[tuple(pred_locs_.transpose())], 
                False
            )
            predictions['scores'] = jnp.where(masked, predictions['scores'], 0)
            # predictions['locations'] = jnp.where(masked[:, None], predictions['locations'], -1)

        return dict(
            detector=x,
            predictions=predictions,
        )



from __future__ import annotations

from functools import partial
from typing import Callable, Optional, Sequence, Iterable, Any

import flax.linen as nn
from flax.linen import initializers
import jax
import jax.numpy as jnp
import numpy as np

from ..typing import Array, ArrayLike

# Inputs are PRNGKey, input shape and dtype.
Initializer = Callable[[jnp.ndarray, Sequence[int], jnp.dtype], jnp.ndarray]

import warnings

def picklable_relu(x):
    return jax.nn.relu(x)

class DefaultUnpicklerMixin:
    """Provides an unpickling method  which will restore
    module instances, even if the model has gained new fields
    or dropped some fields since data was unpickled.
    """
    def __setstate__(self, state):
        try:
            current_fields = self.__class__.__dataclass_fields__
        except AttributeError:
            warnings.warn("DefaultUnpickler used in a class that is not a dataclass")
            current_fields = {}
        for key in list(state.keys()):
            if key not in current_fields:
                if not key[0] == "_":
                    warnings.warn(f"Dropping unused field {key}")
                del state[key]
        for key in current_fields:
            if key not in state:
                if not key == "parent":
                    warnings.warn(f"Adding missing field {key}")
                state[key] = current_fields[key].default

        self.__init__(**state)


class _ChannelAttention(nn.Module):
    squeeze_factor: int = 16
    dtype: Any = None

    @nn.compact
    def __call__(
        self,
        x: ArrayLike,
    ) -> Array:

        orig_shape = x.shape
        se_channels = orig_shape[-1] // self.squeeze_factor

        x_backup = x

        x = x.reshape(-1, orig_shape[-1])
        x = jnp.concatenate([x.max(axis=0), x.mean(axis=0)], axis=-1)

        x = nn.Dense(se_channels, dtype=self.dtype)(x)
        x = jax.nn.relu(x)
        x = nn.Dense(orig_shape[-1], dtype=self.dtype)(x)
        x = jax.nn.sigmoid(x)

        x = x_backup * x

        return x


ChannelAttention = nn.vmap(
    _ChannelAttention, 
    variable_axes={"params":None}, 
    split_rngs={"params":False},
)


class SpatialAttention(nn.Module):
    filter_size: int|tuple[int,...] = 7
    dtype: Any = None

    @nn.compact
    def __call__(
        self,
        x: ArrayLike,
    ) -> Array:

        try:
            conv_filter = len(self.filter_size)
        except:
            conv_filter = (self.filter_size, self.filter_size)
        
        y = jnp.stack([x.max(axis=-1), x.mean(axis=-1)], axis=-1)
        y = nn.Conv(1, conv_filter, dtype=self.dtype)(y)
        y = jax.nn.sigmoid(y)

        y = x * y

        return y


class FFN(nn.Module):
    """A feed-forward block commonly used in transformer"""

    dim: int|None = None
    dropout_rate: float = 0.0
    deterministic: bool = False
    dtype: Any = None

    @nn.compact
    def __call__(self, x, *, deterministic=None):
        deterministic = deterministic or self.deterministic
        orig_dim = x.shape[-1]
        dim = self.dim or orig_dim * 4

        x = nn.LayerNorm(dtype=self.dtype)(x)

        x = nn.Dense(dim, dtype=self.dtype)(x)
        x = nn.gelu(x)
        x = nn.Dropout(self.dropout_rate)(x, deterministic=deterministic)

        x = nn.Dense(orig_dim, dtype=self.dtype)(x)
        x = nn.Dropout(self.dropout_rate)(x, deterministic=deterministic)

        return x


class IdentityLayer(nn.Module):
    """Identity layer, convenient for giving a name to an array."""

    @nn.compact
    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        return x


def get_constant_initializer(constant: float) -> Initializer:
    """Returns an initializer that initializes everything to a given constant."""

    def init_fn(unused_key: jnp.ndarray,  # pytype: disable=annotation-type-mismatch  # jnp-type
                shape: Iterable[int],
                dtype: jnp.dtype = jnp.float32) -> jnp.ndarray:
        return constant * jnp.ones(shape, dtype=dtype)

    return init_fn  # pytype: disable=bad-return-type  # jax-ndarray


class Affine(nn.Module):
    """Affine transformation layer.

    Described in:
    Touvron et al, "ResMLP: Feedforward networks for image classification
    with data-efficient training", 2021.

    Performs an affine transformation on the final dimension of the input tensor.
    """
    bias_init: Initializer = nn.initializers.zeros
    scale_init: Initializer = nn.initializers.ones
    use_bias: bool = True

    @nn.compact
    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        n = x.shape[-1]
        scale = self.param('scale', self.scale_init, (n,))
        if self.use_bias:
            bias = self.param('bias', self.bias_init, (n,))
        else:
            bias = 0.0
        return scale * x + bias


class DropPath(nn.Module):
    """Performs layer-dropout (also known as stochastic depth).

    Described in
    Huang & Sun et al, "Deep Networks with Stochastic Depth", 2016
    https://arxiv.org/abs/1603.09382

    Attributes:
      rate: the layer dropout probability (_not_ the keep rate!).
      deterministic: If false (e.g. in training) the inputs are scaled by `1 / (1
        - rate)` and the layer dropout is applied, whereas if true (e.g. in
        evaluation), no stochastic depth is applied and the inputs are returned as
        is.
    """
    rate: float = 0.0
    deterministic: Optional[bool] = None
    broadcast_dims: Optional[Sequence[int]] = None

    @nn.compact
    def __call__(self,
                 x: jnp.ndarray,
                 deterministic: Optional[bool] = None) -> jnp.ndarray:
        """Applies a stochastic depth mask to the inputs.

        Args:
            x: Input tensor.
            deterministic: If false (e.g. in training) the inputs are scaled by `1 /
                (1 - rate)` and the layer dropout is applied, whereas if true (e.g. in
                evaluation), no stochastic depth is applied and the inputs are returned
                as is.
            broadcast_dims: default assumes the input is a single sample with no batch-axis

        Returns:
            The masked inputs reweighted to preserve mean.
        """
        if self.broadcast_dims is None:
            broadcast_dims = range(x.ndim)
        else:
            broadcast_dims = self.broadcast_dims

        return nn.Dropout(
            rate=self.rate, broadcast_dims=broadcast_dims)(x, deterministic)


class PositionEmbedding1D(nn.Module):
    """ 1D positional embeddings

    Attributes:
      posemb_init: if None, use cosine embedding
      max_len: Maximum possible length for the input. If None, the max_len is
        set to the inputs sequence length.
      max_timescale: time scale for  
      rescale_from: If not None, embeddings are rescaled from this length.
    """
    posemb_init: Initializer|None = nn.initializers.normal(stddev=0.02)
    max_len: int|None = None
    max_timescale: float = 1.0e4
    rescale_from: int|None = None

    @nn.compact
    def __call__(self, input: tuple|ArrayLike) -> Array:
        """
        Args:
          inputs: Input data that needs embedding or shape tuple

        Returns:
          Output: `(Sequence_length, inputs_dim)`.
        """
        if isinstance(input, tuple):
            input_shape = input
        else:
            input_shape = input.shape

        assert len(input_shape) >= 2, f"invalid input shape {input_shape}"
        length, dim = input_shape[-2:]
        max_len = self.max_len or length

        if self.rescale_from: 
            embedding_length = self.rescale_from
        else:
            embedding_length = max_len

        if self.posemb_init is not None:
            pos_emb = self.param("pos_emb", self.posemb_init, (embedding_length, dim))
        else:
            pos_emb = np.zeros((embedding_length, dim), dtype=np.float32)
            position = np.arange(embedding_length)[:, np.newaxis]
            div_term = np.exp(
                np.arange(0, dim, 2) / dim * np.log(self.max_timescale)
            )
            pos_emb[:, 0::2] = np.sin(position / div_term)
            pos_emb[:, 1::2] = np.cos(position / div_term)
            pos_emb = jnp.asarray(pos_emb)

        if self.rescale_from:
            pos_emb = jax.image.resize(
                pos_emb, (max_len, dim), method='bilinear', antialias=False
            )

        pos_emb = pos_emb[:length, :]

        assert pos_emb.shape == (length, dim)

        return pos_emb


class PositionEmbedding2D(nn.Module):
    """2D cosine positional embeddings

    Attributes:
      rescale_from: If not None, embeddings are rescaled from this shape.
      posemb_init: Positional embedding initializer.
    """
    posemb_init: Initializer = nn.initializers.normal(stddev=0.02)
    rescale_from: tuple[int, int]|None = None

    @nn.compact
    def __call__(self, input: tuple|ArrayLike) -> Array:
        """
        Args:
          inputs: Input data that needs embedding or shape tuple

        Returns:
          Output: `(h, w, inputs_dim)`.
        """
        if isinstance(input, tuple):
            input_shape = input
        else:
            input_shape = input.shape

        assert len(input_shape) >= 3, f"invalid input shape {input_shape}"
        h, w, c = input_shape[-3:]

        if self.rescale_from:  # `[h, w, c]`
            embedding_h, embedding_w = self.rescale_from[0], self.rescale_from[1]
        else:
            embedding_h, embedding_w = h, w

        row_pos_embed = self.param('row_pos_embedding', self.posemb_init, (embedding_w, c // 2))
        col_pos_embed = self.param('col_pos_embedding', self.posemb_init, (embedding_h, c // 2))

        # To `[h, w, c//2]`.
        x_pos_emb = jnp.tile(
            jnp.expand_dims(row_pos_embed, axis=0), (embedding_h, 1, 1)
        )
        y_pos_emb = jnp.tile(
            jnp.expand_dims(col_pos_embed, axis=1), (1, embedding_w, 1)
        )

        # To `[h, w, c]`.
        pos = jnp.concatenate((x_pos_emb, y_pos_emb), axis=-1)

        if self.rescale_from:
            pos = jax.image.resize(
                pos, (h, w, c), method='bilinear', antialias=False
            )

        return pos


class GRN(nn.Module):
    epsilon: float = 1e-6

    @nn.compact
    def __call__(self, x):
        dim = x.shape[-1]
        gamma = self.param("gamms", nn.initializers.zeros, (1, 1, 1, dim))
        beta = self.param("beta", nn.initializers.zeros, (1, 1, 1, dim))

        x_ = jnp.array(x, "float32")
        mu2 = jax.lax.square(jnp.abs(x_)).mean(axis=(1,2), keepdims=True)
        mu2 = jnp.maximum(mu2, 1e-6)
        Gx = jax.lax.sqrt(mu2).astype(x.dtype)
        Nx = Gx / (Gx.mean(axis=-1, keepdims=True) + self.epsilon)

        return gamma * (x * Nx) + beta + x



from __future__ import annotations

import jax
import optax
from ml_collections import ConfigDict

from ..ops import match_and_replace, distance_similarity
from ..typing import ArrayLike
from ..utils import deep_update

jnp = jax.numpy

def _compute_losses(module, gt_locs, mask, decoded, config):
    gamma = config.get("detection_loss_gamma", 2)
    delta = config.get("detection_loss_delta", 4.0)

    pred_locs = decoded['pred_locs']
    ref_locs = decoded['ref_locs']
    pred_logits = decoded['logits']

    if mask is not None:
        mask = mask[tuple(ref_locs.transpose().astype(int))]
    else:
        mask = jnp.array([True])

    indices, regr_mask = _assign_label(module, gt_locs, decoded, config)
    gt_regr = gt_locs[indices]

    cls_loss = optax.sigmoid_focal_loss(
        pred_logits, regr_mask, gamma=gamma,
    ).mean(where=mask)

    regr_loss = optax.l2_loss(
        pred_locs/delta, 
        gt_regr/delta,
    ).mean(axis=-1).sum(where = regr_mask)
    regr_loss /= jnp.count_nonzero(regr_mask) + 1e-6

    return dict(
        lpn_localization_loss=regr_loss,
        lpn_detection_loss=cls_loss,
    )


def _assign_label(module, gt_locs, decoded, config):
    locs = decoded['pred_locs']
    ref_locs = decoded['ref_locs']
    assert locs.ndim == 2 and locs.shape[-1] in (2, 3), f"{locs.shape}"
    assert gt_locs.shape[-1] == locs.shape[-1]

    assert ref_locs.ndim == 2 and ref_locs.shape[-1] in (2, 3), f"{ref_locs.shape}"

    detection_roi = config.get("detection_roi", 8)
    n_labels_max = config.get("n_labels_max", 25)
    n_labels_min = config.get("n_labels_min", 4)
    similarity_score_scaling = config.get("similarity_score_scaling", 4)

    distances = gt_locs - ref_locs[:, None, :] #[M, N, 2/3]
    init_mask = (jnp.abs(distances) < detection_roi).all(axis=-1) # [M. N]

    sm = jnp.where(init_mask, distance_similarity(locs, gt_locs), 0) #[M, N]

    M, N = sm.shape

    sm_topk, idx = jax.lax.approx_max_k(sm.transpose(), n_labels_max) # [N, 25]
    sm_topk *= similarity_score_scaling

    k = jnp.sqrt(jnp.clip(sm_topk, 1e-16, 1)).sum(axis=1)
    k = jnp.clip(k, n_labels_min, n_labels_max)
    idx = jnp.where(
        (jnp.arange(n_labels_max) < k[:, None]) & (gt_locs[:, -1:] >= 0),
        idx, M,
    )

    indices = jnp.zeros(M, int) - 1 
    indices = indices.at[idx].set(jnp.arange(N)[:, None]) # this relies on Jax's oob behavior
    regr_mask = indices >= 0

    return indices, regr_mask


def train_fn(
    module,
    image: ArrayLike,
    gt_locations: ArrayLike,
    video_refs: tuple|None = None,
    image_mask: ArrayLike|None = None,
    config: ConfigDict = ConfigDict(),
) -> dict:
    image = jnp.asarray(image)
    gt_locations = jnp.asarray(gt_locations)

    assert image.ndim in (3, 4) and image.shape[-1] == 3, f"wrong img shape {image.shape}"
    assert gt_locations.ndim == 2 and gt_locations.shape[1] in (2,3), f"wrong location data shape {gt_locations.shape}"

    max_proposal_offset = config.get("max_proposal_offset", 12)
    max_training_instances = config.get("max_training_instances", 256)

    is_3d = image.ndim == 4

    if image_mask is not None:
        image_mask = jnp.broadcast_to(image_mask, image.shape[:-1])

    x_det, x_seg = module.get_features(image, video_refs, deterministic=is_3d)

    if not is_3d:
        assert x_det.ndim == 3
        outputs = module.detector(x_det, image_mask)
    else:
        assert x_det.ndim == 4
        outputs = module.detector_3d(x_det, image_mask)

    outputs['losses'] = _compute_losses(module, gt_locations, image_mask, outputs['detector'], config)

    if not is_3d and module.segmentor is not None:
        max_proposal_offset = config.get("max_proposal_offset", 12)
        max_training_instances = config.get("max_training_instances", 256)

        match_th = 1 / max_proposal_offset / max_proposal_offset
        seg_locs = match_and_replace(
            gt_locations,  outputs["predictions"]["locations"], match_th
        )
        seg_locs = seg_locs[:max_training_instances]

        segmentor_out = module.segmentor(x_seg, seg_locs)

        outputs = deep_update(outputs, segmentor_out)

    elif is_3d and module.segmentor_3d is not None:
        max_proposal_offset = config.get("max_proposal_offset", 12)
        max_training_instances = config.get("max_training_instances", 32)

        match_th = 1 / max_proposal_offset / max_proposal_offset
        seg_locs = match_and_replace(
            gt_locations,  outputs["predictions"]["locations"], match_th
        )
        seg_locs = seg_locs[:max_training_instances]

        segmentor_out = module.segmentor_3d(x_seg, seg_locs)

        outputs = deep_update(outputs, segmentor_out)
    
    return outputs



import logging
import threading
import traceback
from concurrent import futures
from pathlib import Path

import grpc
import jax
import numpy as np
import typer

from . import proto
from .common import get_dtype, decode_image, TokenValidationInterceptor

app = typer.Typer(pretty_exceptions_enable=False)

_MAX_MSG_SIZE=1024*1024*128
_TARGET_CELL_SIZE=32


def process_input(request: proto.DetectionRequest):
    pixels = request.image_data.pixels
    settings = request.detection_settings

    image = decode_image(pixels)

    physical_size = np.array([
        pixels.physical_size_z or pixels.physical_size_x, # one might set xy but not z
        pixels.physical_size_y, 
        pixels.physical_size_x,
    ], dtype="float")
    if (physical_size == 0).any():
        physical_size[:] = 1.0

    if settings.HasField("cell_diameter_hint"):
        scaling = _TARGET_CELL_SIZE / settings.cell_diameter_hint * physical_size

    else:
        if physical_size[1] != physical_size[2]:
            raise ValueError("Scaling hint provided, but pixel is not isometric")

        scaling = np.array([ settings.scaling_hint or 1.0 ] * 3, dtype="float")
        scaling[0] *= physical_size[0] / physical_size[1] 
    
    logging.info(f"Requested rescaling factor is {scaling}")

    shape_hint = tuple( np.round(scaling * image.shape[:3]).astype(int) )

    if image.shape[0] == 1: # 2D
        image = image.squeeze(0)
        shape_hint = shape_hint[1:]

    kwargs = dict(
        reshape_to = shape_hint,
        score_threshold = settings.min_score or 0.4,
        min_area = settings.min_cell_area,
        nms_iou = settings.nms_iou,
        segmentation_threshold = settings.segmentation_threshold or 0.5,
    )

    return image, kwargs


def process_result(preds, image) -> proto.DetectionResponse:
    response = proto.DetectionResponse()

    if image.ndim == 3: # returns polygon

        for contour, score in zip(preds["pred_contours"], preds["pred_scores"]):
            if len(contour) == 0:
                continue

            scored_roi = proto.ScoredROI(
                score = score,
                roi = proto.ROI(
                    polygon = proto.Polygon(points = [proto.Point(x=p[0], y=p[1]) for p in contour]),
                )
            )

            response.detections.append(scored_roi)

    else: # 3d returns Mesh
        for mesh, score in zip(preds["pred_contours"], preds["pred_scores"]):
            scored_roi = proto.ScoredROI(
                score = score,
                roi = proto.ROI(
                    mesh = proto.Mesh(
                        verts = [proto.Point(z=v[0], y=v[1], x=v[2]) for v in mesh['verts']],
                        faces = [proto.Mesh.Face(p1=p[0], p2=p[1], p3=p[2]) for p in mesh['faces']],
                    ),
                )
            )

            response.detections.append(scored_roi)

    return response


class LacssServicer(proto.LacssServicer):

    def __init__(self, model, max_image_size, max_image_size_3d):
        self.model = model
        self.max_image_size = max_image_size
        self.max_image_size_3d = max_image_size_3d

        self._lock = threading.RLock()

    def RunDetection(self, request, context):
        with self._lock:

            logging.info(f"Received message of size {request.ByteSize()}")

            try:
                image, kwargs = process_input(request)

                reshape_to = kwargs["reshape_to"]

                if image.ndim == 4:
                    is_img_too_big = max(reshape_to) > self.max_image_size_3d
                else:
                    is_img_too_big = max(reshape_to) > self.max_image_size

                if is_img_too_big:
                    raise ValueError("image size exeeds limit")

                logging.info(f"received image {image.shape}")

                # dont' reshape image if the request is almost same as the orginal 
                rel_diff = np.abs(np.array(reshape_to) / image.shape[:-1] - 1)
                if (rel_diff < 0.1).all():
                    kwargs['reshape_to'] = None

                preds = self.model.predict(
                    image, output_type="contour", **kwargs,
                )

                response = process_result(preds, image)

                logging.info(f"Reply with message of size {response.ByteSize()}")

                return response
            
            except ValueError as e:
                
                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(grpc.StatusCode.INVALID_ARGUMENT, repr(e))

            except Exception as e:

                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}")



    def RunDetectionStream(self, request_iterator, context):
        with self._lock:
            request = proto.DetectionRequest()

            for next_request in request_iterator:

                if next_request.image_data.HasField("pixels"):
                    request.image_data.pixels.CopyFrom(next_request.image_data.pixels)

                if next_request.image_data.HasField("image_annotation"):
                    request.image_data.image_annotation.CopyFrom(next_request.image_data.image_annotation)
                
                if next_request.HasField("detection_settings"):
                    request.detection_settings.CopyFrom(next_request.detection_settings)
                
                if request.image_data.HasField("pixels"):
                    yield self.RunDetection(request, context)


def get_predictor(modelpath):
    from .predict import Predictor

    model = Predictor(modelpath)

    logging.info(f"lacss_server: loaded model from {modelpath}")

    model.module.detector.max_output = 512  # FIXME good default?
    model.module.detector.min_score = 0.05

    logging.debug(f"lacss_server: precompile for most common shapes")

    _ = model.predict(np.ones([512,512,3]), output_type="contour")
    _ = model.predict(np.ones([1088,1088,3]), output_type="contour")
    _ = model.predict(np.ones([256,256,256,3]), output_type="contour")

    return model


def show_urls():
    from . import model_urls
    
    print("Pretrained model files:")
    print("==============================")
    for k, v in model_urls.items():
        print(f"{k}: {v}")
    print()


@app.command()
def main(
    modelpath: Path|None = None,
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool|None = None,
    debug: bool = False,
    compression: bool = True,
    max_image_size: int = 1088,
    max_image_size_3d: int = 512,
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    if modelpath is None:
        show_urls()
        return

    print ("server starting ...")

    model = get_predictor(modelpath)

    logging.info(f"lacss_server: default backend is {jax.default_backend()}")

    if jax.default_backend() == "cpu":
        logging.warning(
            f"lacss_server: WARNING: No GPU configuration. This might be very slow ..."
        )

    if token is None:
        token = not local
    if token:
        import secrets

        token_str = secrets.token_urlsafe(64)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token_str}")
        print("=======================================================================")
        print()
    else:
        token_str = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip if compression else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token_str),),
        options=(("grpc.max_receive_message_length", _MAX_MSG_SIZE),),
    )

    proto.add_LacssServicer_to_server(
        LacssServicer(
            model,
            max_image_size,
            max_image_size_3d,
        ), 
        server,
    )

    if local:
        server.add_secure_port(f"127.0.0.1:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logging.info(f"lacss_server: listening on port {port}")

    print ("server starting ... ready")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    # jax.config.update("jax_compilation_cache_dir", "jax_cache")
    # jax.config.update("jax_persistent_cache_min_entry_size_bytes", -1)
    # jax.config.update("jax_persistent_cache_min_compile_time_secs", 5)
    app()
